import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.matrix_vector_multip import mv
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('matrix_vector_multip', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 20):  # Adjust the range for different sizes
            M = 128 * i
            N = 128 * (i - 1)  # Example: N is half of M
            matrix = torch.rand((N, M), dtype=torch.float32)
            vector = torch.rand((M,), dtype=torch.float32)
            self.input_tensors.append((matrix, vector))

    def to_cuda(self, input_tensor):
        matrix, vector = input_tensor
        return (matrix.cuda(), vector.cuda())

    def call_op(self, input_tensor):
        matrix, vector = input_tensor
        return mv(matrix, vector)

    def get_gbps(self, input_tensor, runtime):
        matrix, vector = input_tensor
        total_bytes = (matrix.numel() + vector.numel() + matrix.size(0)) * matrix.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        matrix, vector = input_tensor
        N, M = matrix.shape
        FLOPS = 2 * N * M  # Each element in the output involves M multiplications and M-1 additions
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def rmsnorm_triton(x_ptr, rms_w_ptr, out_ptr,
                   stride_x_batch, stride_x_m, stride_x_k,
                   stride_rms_w,
                   stride_out_batch, stride_out_m, stride_out_k,
                   N_SIZE: tl.constexpr, eps: tl.constexpr, BLOCK_N_SIZE: tl.constexpr):
    pid_batch = tl.program_id(0)
    pid_m = tl.program_id(1)

    # parallel at m dimension
    offset_m = pid_batch * stride_x_batch + pid_m * stride_x_m
    block_n_size = tl.arange(0, BLOCK_N_SIZE)
    var = tl.zeros((BLOCK_N_SIZE,), tl.float32)
    # parallel between blocks
    for block_n_strart_ptr in range(0, N_SIZE, BLOCK_N_SIZE):
        offset_n = block_n_strart_ptr + block_n_size
        x_ptr_mask = offset_n < N_SIZE
        x = tl.load(x_ptr + offset_m + offset_n * stride_x_k, mask=x_ptr_mask, other=0.)  # careful stride_x_k
        xf = x.to(tl.float32)
        var += xf*xf
    var = tl.sum(var, axis=0) / N_SIZE  # reduce between wrap
    std = tl.sqrt(var + eps)

    for block_n_strart_ptr in range(0, N_SIZE, BLOCK_N_SIZE):
        offset_n = block_n_strart_ptr + block_n_size
        x_ptr_mask = offset_n < N_SIZE

        rms_w_offset = tl.load(rms_w_ptr + offset_n * stride_rms_w, mask=x_ptr_mask)
        x = tl.load(x_ptr + offset_m + offset_n * stride_x_k, mask=x_ptr_mask, other=0.)

        x_new = x / std
        out = x_new * rms_w_offset
        out_offset = pid_batch * stride_out_batch + pid_m * stride_out_m + offset_n * stride_out_k
        tl.store(out_ptr + out_offset, out, mask=x_ptr_mask)


def rmsnorm_wrapper(x, rms_weights, eps=1e-6):
    batch, M, K = x.shape
    out = torch.empty_like(x)
    rmsnorm_triton[(batch, M,)](x, rms_weights, out,
                                *x.stride(),
                                *rms_weights.stride(),  # 1
                                *out.stride(),
                                N_SIZE=K, eps=eps, BLOCK_N_SIZE=4096,
                                num_warps=16
                                )
    return out




##################################################################################################################################################


def test_rmsnorm():
    # Define the input tensor x with shape (batch, M, K)
    batch = 2
    M = 3
    K = 4096
    x = torch.randn((batch, M, K), dtype=torch.float16, device="cuda")

    # Define the rms_weights tensor with shape (K,)
    rms_weights = torch.randn((K,), dtype=torch.float16, device="cuda")

    # Dictionary to store the results of different test cases
    results = {}

    # Test case 1
    out1 = rmsnorm_wrapper(x, rms_weights)
    results['test_case_1'] = out1.cpu()

    # Additional test cases for branch coverage

    # Test case 2: Different batch size
    batch = 4
    x = torch.randn((batch, M, K), dtype=torch.float16, device="cuda")
    out2 = rmsnorm_wrapper(x, rms_weights)
    results['test_case_2'] = out2.cpu()

    # Test case 3: Different M size
    M = 5
    x = torch.randn((batch, M, K), dtype=torch.float16, device="cuda")
    out3 = rmsnorm_wrapper(x, rms_weights)
    results['test_case_3'] = out3.cpu()

    # Test case 4: Larger K size
    K = 8192
    rms_weights = torch.randn((K,), dtype=torch.float16, device="cuda")
    x = torch.randn((batch, M, K), dtype=torch.float16, device="cuda")
    out4 = rmsnorm_wrapper(x, rms_weights)
    results['test_case_4'] = out4.cpu()

    return results

# Execute the test function
result_gold = test_rmsnorm()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.fused_activation import fused_add_mul_activation_torch
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('fused_activation', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(12, 30):
            size = 2 ** i
            in_out_tensor = torch.rand(size, dtype=torch.float16)
            bias = torch.rand(size // 4, dtype=torch.float16)  # Assuming num_weights is size // 4
            in_tensor = torch.rand(size, dtype=torch.float16)
            self.input_tensors.append((in_out_tensor, bias, in_tensor))

    def to_cuda(self, input_tensor):
        in_out_tensor, bias, in_tensor = input_tensor
        return (in_out_tensor.cuda(), bias.cuda(), in_tensor.cuda())

    def call_op(self, input_tensor):
        in_out_tensor, bias, in_tensor = input_tensor
        return fused_add_mul_activation_torch(in_out_tensor, bias, in_tensor)

    def get_gbps(self, input_tensor, runtime):
        in_out_tensor, bias, in_tensor = input_tensor
        total_bytes = (in_out_tensor.numel() + bias.numel() + in_tensor.numel()) * in_out_tensor.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        in_out_tensor, bias, in_tensor = input_tensor
        FLOPS = 2 * in_out_tensor.numel()  # Assuming 2 operations per element (add and multiply)
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.softmax_reducev import token_softmax_reducev_fwd
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('softmax_reducev', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(5, 12):  # Example sizes, adjust as needed
            batch_size = 2 ** i
            seq_len = 128  # Example sequence length
            d_model = 64  # Example model dimension
            logics = torch.rand((batch_size, seq_len), dtype=torch.float32)
            v = torch.rand((batch_size, seq_len, d_model), dtype=torch.float32)
            o = torch.empty((batch_size, seq_len, d_model), dtype=torch.float32)
            b_loc = torch.randint(0, seq_len, (batch_size, seq_len), dtype=torch.int32)
            b_start_loc = torch.zeros(batch_size, dtype=torch.int32)
            b_seq_len = torch.full((batch_size,), seq_len, dtype=torch.int32)
            max_input_len = seq_len
            other_kv_index = 0  # Example value
            self.input_tensors.append((logics, v, o, b_loc, b_start_loc, b_seq_len, max_input_len, other_kv_index))

    def to_cuda(self, input_tensor):
        # return tuple(tensor.cuda() for tensor in input_tensor)
        logics, v, o, b_loc, b_start_loc, b_seq_len, max_input_len, other_kv_index = input_tensor
        return (logics.cuda(), v.cuda(), o.cuda(), b_loc.cuda(), b_start_loc.cuda(), b_seq_len.cuda(), max_input_len, other_kv_index)

    def call_op(self, input_tensor):
        logics, v, o, b_loc, b_start_loc, b_seq_len, max_input_len, other_kv_index = input_tensor
        return token_softmax_reducev_fwd(logics, v, o, b_loc, b_start_loc, b_seq_len, max_input_len, other_kv_index)

    def get_gbps(self, input_tensor, runtime):
        logics, v, o, _, _, _, _, _ = input_tensor
        total_bytes = (logics.numel() + v.numel() + o.numel()) * logics.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        logics, v, _, _, _, _, _, _ = input_tensor
        FLOPS = 2 * logics.numel() * v.shape[-1]  # Assuming 2 FLOPS per element for softmax and reduce
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl


@triton.jit
def fused_add_mul_activation_kernel(x_ptr, bias_ptr, in_ptr,
                                    num_weights: tl.constexpr,
                                    xnumel: tl.constexpr,
                                    multiplier: tl.constexpr,
                                    activation: tl.constexpr,
                                    BLOCK_SIZE: tl.constexpr):
    xoffset = tl.program_id(0) * BLOCK_SIZE
    index = xoffset + tl.arange(0, BLOCK_SIZE)[:]
    mask = index < xnumel
    bias_index = index % num_weights
    tmp0 = tl.load(x_ptr + index, mask)
    tmp1 = tl.load(bias_ptr + bias_index, mask, eviction_policy='evict_last')
    tmp3 = tl.load(in_ptr + index, mask)
    activ_input = multiplier * tmp3 + tmp0 + tmp1
    if activation == "sigmoid":
        ma_result = tl.sigmoid(activ_input)
        # option 2 - calculate sigmoid using exp
        # ma_result = 1.0 / (1.0 + tl.exp(-sigmoid_input))
        # option 3: fast sigmoid - inaccurate but faster
        # ma_result = 1.0 / (1.0 + tl.abs(sigmoid_input))
    elif activation == "relu":
        ma_result = tl.maximum(0, activ_input)

    tl.store(x_ptr + index, ma_result, mask)


def fused_add_mul_activation_torch(in_out_tensor: torch.Tensor, bias: torch.Tensor,
                                   in_tensor: torch.Tensor) -> torch.Tensor:
    # print("calling fused_add_mul_relu_torch")
    grid = lambda meta: (triton.cdiv(in_out_tensor.numel(), meta['BLOCK_SIZE']),)
    BLOCK_SIZE = min(2048, in_out_tensor.numel())
    fused_add_mul_activation_kernel[grid](in_out_tensor, bias, in_tensor,
                                          bias.numel(),
                                          in_out_tensor.numel(),
                                          multiplier=0.5,
                                          activation="sigmoid",
                                          BLOCK_SIZE=BLOCK_SIZE)
    return in_out_tensor




##################################################################################################################################################


def test_fused_add_mul_activation():
    # 输入张量形状
    num_elements = 8192  # 总元素数量
    num_weights = 64     # 偏置的数量

    # 创建输入张量
    in_out_tensor = torch.randn(num_elements, dtype=torch.float32, device='cuda')  # 输入输出张量
    bias = torch.randn(num_weights, dtype=torch.float32, device='cuda')  # 偏置
    in_tensor = torch.randn(num_elements, dtype=torch.float32, device='cuda')  # 额外输入张量

    # 分支1: activation="sigmoid"
    result_sigmoid = fused_add_mul_activation_torch(in_out_tensor.clone(), bias, in_tensor)

    # 分支2: activation="relu"
    grid = lambda meta: (triton.cdiv(in_out_tensor.numel(), meta['BLOCK_SIZE']),)
    BLOCK_SIZE = min(2048, in_out_tensor.numel())
    fused_add_mul_activation_kernel[grid](in_out_tensor, bias, in_tensor,
                                          bias.numel(),
                                          in_out_tensor.numel(),
                                          multiplier=0.5,
                                          activation="relu",
                                          BLOCK_SIZE=BLOCK_SIZE)
    result_relu = in_out_tensor.clone()

    # 分支覆盖率【2/4】
    results = {
        "test_case_1": result_sigmoid[:10].cpu().numpy(),
        "test_case_2": result_relu[:10].cpu().numpy()
    }
    return results

result_gold = test_fused_add_mul_activation()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

# Triton kernel for forward token softmax
@triton.jit
def _fwd_kernel_token_softmax(
    Logics, B_Start_Loc, B_Seqlen,
    Prob_Out,
    stride_logic_h, stride_logic_bs,
    stride_prob_h, stride_prob_bs,
    BLOCK_SIZE: tl.constexpr
):
    cur_batch = tl.program_id(0)
    cur_head = tl.program_id(1)

    col_offsets = tl.arange(0, BLOCK_SIZE)
    cur_batch_seq_len = tl.load(B_Seqlen + cur_batch)
    cur_batch_in_all_start_index = tl.load(B_Start_Loc + cur_batch)

    row = tl.load(Logics + cur_head * stride_logic_h + (cur_batch_in_all_start_index + col_offsets) * stride_logic_bs,
                  mask=col_offsets < cur_batch_seq_len, other=-float('inf')).to(tl.float32)

    row_minus_max = row - tl.max(row, axis=0)
    numerator = tl.exp(row_minus_max)
    denominator = tl.sum(numerator, axis=0)
    softmax_output = numerator / denominator

    tl.store(Prob_Out + cur_head * stride_prob_h + (cur_batch_in_all_start_index + col_offsets)
             * stride_prob_bs, softmax_output, mask=col_offsets < cur_batch_seq_len)
    return

# Function to launch the Triton kernel
@torch.no_grad()
def token_softmax_fwd(Logics, B_Start_Loc, B_Seqlen, Prob_Out, max_input_len):
    BLOCK_SIZE = triton.next_power_of_2(max_input_len)
    batch, head_num = B_Start_Loc.shape[0], Logics.shape[0]

    num_warps = 4
    if BLOCK_SIZE >= 2048:
        num_warps = 8
    if BLOCK_SIZE >= 4096:
        num_warps = 16

    _fwd_kernel_token_softmax[(batch, head_num)](
        Logics, B_Start_Loc, B_Seqlen,
        Prob_Out,
        Logics.stride(0), Logics.stride(1),
        Prob_Out.stride(0), Prob_Out.stride(1),
        num_warps=num_warps,
        BLOCK_SIZE=BLOCK_SIZE,
    )
    return




##################################################################################################################################################


import torch

# Define the test function
def test_token_softmax_fwd():
    results = {}

    # Test case 1: Small input size
    batch_size = 2
    head_num = 2
    max_input_len = 8

    # Create random input tensors
    Logics = torch.randn((head_num, batch_size * max_input_len), dtype=torch.float32, device='cuda')
    B_Start_Loc = torch.tensor([0, max_input_len], dtype=torch.int32, device='cuda')
    B_Seqlen = torch.tensor([max_input_len, max_input_len], dtype=torch.int32, device='cuda')
    Prob_Out = torch.empty_like(Logics)

    # Call the Triton softmax function
    token_softmax_fwd(Logics, B_Start_Loc, B_Seqlen, Prob_Out, max_input_len)

    # Store the output
    results['test_case_1'] = Prob_Out.clone()

    # Test case 2: Larger input size
    batch_size = 1
    head_num = 1
    max_input_len = 16

    # Create random input tensors
    Logics = torch.randn((head_num, batch_size * max_input_len), dtype=torch.float32, device='cuda')
    B_Start_Loc = torch.tensor([0], dtype=torch.int32, device='cuda')
    B_Seqlen = torch.tensor([max_input_len], dtype=torch.int32, device='cuda')
    Prob_Out = torch.empty_like(Logics)

    # Call the Triton softmax function
    token_softmax_fwd(Logics, B_Start_Loc, B_Seqlen, Prob_Out, max_input_len)

    # Store the output
    results['test_case_2'] = Prob_Out.clone()

    return results

# Run the test function
result_gold = test_token_softmax_fwd()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.adam_update_triton import update_fn
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('adam_update_triton', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(12, 28):
            size = 2 ** i
            p = torch.rand(size, dtype=torch.float32)
            grad = torch.rand(size, dtype=torch.float32)
            exp_avg = torch.rand(size, dtype=torch.float32)
            self.input_tensors.append((p, grad, exp_avg))

    def to_cuda(self, input_tensor):
        p, grad, exp_avg = input_tensor
        return (p.cuda(), grad.cuda(), exp_avg.cuda())

    def call_op(self, input_tensor):
        p, grad, exp_avg = input_tensor
        lr = 0.001
        wd = 0.01
        beta1 = 0.9
        beta2 = 0.999
        update_fn(p, grad, exp_avg, lr, wd, beta1, beta2)

    def get_gbps(self, input_tensor, runtime):
        p, grad, exp_avg = input_tensor
        total_bytes = 3 * p.numel() * p.element_size()  # p, grad, exp_avg
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        p, grad, exp_avg = input_tensor
        FLOPS = 2 * p.numel()  # Simplified estimation
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config()
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def _triton_rope(
    q_ptr,
    q_row_stride,
    k_ptr,
    k_row_stride,
    cos,
    cos_row_stride,
    sin,
    sin_row_stride,
    sl,
    bs: tl.constexpr,
    n_qh: tl.constexpr,
    n_kh: tl.constexpr,
    hd: tl.constexpr,
    pad_n_qh: tl.constexpr,
    pad_n_kh: tl.constexpr,
    pad_hd: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    BACKWARD_PASS: tl.constexpr = False,
):
    pid = tl.program_id(0)

    q_ptr = q_ptr + pid * q_row_stride
    k_ptr = k_ptr + pid * k_row_stride

    cos_row_idx = pid % (sl)
    cos = cos + cos_row_idx * cos_row_stride
    sin = sin + cos_row_idx * sin_row_stride
    cos_offsets = tl.arange(0, pad_hd // 2)
    cos_mask = cos_offsets < hd // 2
    cos_row = tl.load(cos + cos_offsets, mask=cos_mask, other=0)
    sin_row = tl.load(sin + cos_offsets, mask=cos_mask, other=0)

    first_half_q_offsets = tl.arange(0, pad_n_qh)[:, None] * hd + tl.arange(0, pad_hd // 2)[None, :]
    first_half_k_offsets = tl.arange(0, pad_n_kh)[:, None] * hd + tl.arange(0, pad_hd // 2)[None, :]
    first_q_mask = (tl.arange(0, pad_n_qh)[:, None] < n_qh) & (tl.arange(0, pad_hd // 2)[None, :] < hd // 2)
    first_k_mask = (tl.arange(0, pad_n_kh)[:, None] < n_kh) & (tl.arange(0, pad_hd // 2)[None, :] < hd // 2)
    q_tile_1 = tl.load(q_ptr + first_half_q_offsets, mask=first_q_mask, other=0).to(sin_row.dtype)
    k_tile_1 = tl.load(k_ptr + first_half_k_offsets, mask=first_k_mask, other=0).to(sin_row.dtype)

    second_half_q_offsets = first_half_q_offsets + (hd // 2)
    second_half_k_offsets = first_half_k_offsets + (hd // 2)
    second_q_mask = first_q_mask
    second_k_mask = first_k_mask
    q_tile_2 = tl.load(q_ptr + second_half_q_offsets, mask=second_q_mask, other=0).to(sin_row.dtype)
    k_tile_2 = tl.load(k_ptr + second_half_k_offsets, mask=second_k_mask, other=0).to(sin_row.dtype)

    if not BACKWARD_PASS:
        new_q_tile_1 = q_tile_1 * cos_row - q_tile_2 * sin_row
        tl.store(q_ptr + first_half_q_offsets, new_q_tile_1, mask=first_q_mask)
        new_q_tile_2 = q_tile_2 * cos_row + q_tile_1 * sin_row
        tl.store(q_ptr + second_half_q_offsets, new_q_tile_2, mask=second_q_mask)

        new_k_tile_1 = k_tile_1 * cos_row - k_tile_2 * sin_row
        tl.store(k_ptr + first_half_k_offsets, new_k_tile_1, mask=first_k_mask)
        new_k_tile_2 = k_tile_2 * cos_row + k_tile_1 * sin_row
        tl.store(k_ptr + second_half_k_offsets, new_k_tile_2, mask=second_k_mask)
    else:
        new_q_tile_1 = q_tile_1 * cos_row + q_tile_2 * sin_row
        tl.store(q_ptr + first_half_q_offsets, new_q_tile_1, mask=first_q_mask)
        new_q_tile_2 = q_tile_2 * cos_row - q_tile_1 * sin_row
        tl.store(q_ptr + second_half_q_offsets, new_q_tile_2, mask=second_q_mask)

        new_k_tile_1 = k_tile_1 * cos_row + k_tile_2 * sin_row
        tl.store(k_ptr + first_half_k_offsets, new_k_tile_1, mask=first_k_mask)
        new_k_tile_2 = k_tile_2 * cos_row - k_tile_1 * sin_row
        tl.store(k_ptr + second_half_k_offsets, new_k_tile_2, mask=second_k_mask)


def rope_forward(q, k, cos, sin):
    q = q.transpose(1, 2)
    k = k.transpose(1, 2)

    batch_size, seq_len, n_q_head, head_dim = q.shape
    n_kv_head = k.shape[2]
    pad_hd = triton.next_power_of_2(head_dim)
    pad_n_q_head = triton.next_power_of_2(n_q_head)
    pad_n_kv_head = triton.next_power_of_2(n_kv_head)
    BLOCK_SIZE = max(pad_n_q_head, pad_n_kv_head)

    n_row = batch_size * seq_len

    q = q.contiguous()
    k = k.contiguous()
    cos = cos.contiguous()
    sin = sin.contiguous()

    _triton_rope[(n_row,)](
        q,
        q.stride(1),
        k,
        k.stride(1),
        cos,
        cos.stride(-2),
        sin,
        sin.stride(-2),
        seq_len,
        batch_size,
        n_q_head,
        n_kv_head,
        head_dim,
        pad_n_q_head,
        pad_n_kv_head,
        pad_hd,
        BLOCK_SIZE=BLOCK_SIZE,
        BACKWARD_PASS=False,
    )
    return q.transpose(1, 2), k.transpose(1, 2), cos, sin



##################################################################################################################################################


import torch

def test_rope_forward():
    # Define the test parameters
    batch_size = 2
    seq_len = 4
    n_q_head = 8
    n_kv_head = 8
    head_dim = 16

    # Create random input tensors
    q = torch.randn(batch_size, n_q_head, seq_len, head_dim, dtype=torch.float32, device='cuda')
    k = torch.randn(batch_size, n_kv_head, seq_len, head_dim, dtype=torch.float32, device='cuda')
    cos = torch.randn(seq_len, head_dim // 2, dtype=torch.float32, device='cuda')
    sin = torch.randn(seq_len, head_dim // 2, dtype=torch.float32, device='cuda')

    # Dictionary to store results for each test case
    results = {}

    # Test case 1: Forward pass
    q_out_1, k_out_1, cos_out_1, sin_out_1 = rope_forward(q, k, cos, sin)
    results['test_case_1'] = (q_out_1, k_out_1, cos_out_1, sin_out_1)

    # Test case 2: Backward pass
    q_out_2, k_out_2, cos_out_2, sin_out_2 = rope_forward(q, k, cos, sin)
    results['test_case_2'] = (q_out_2, k_out_2, cos_out_2, sin_out_2)

    return results

result_gold = test_rope_forward()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def matmul_tma_load_store(
        a_ptr, b_ptr, c_ptr,
        M, N, K,
        stride_am, stride_ak,
        stride_bk, stride_bn,
        stride_cm, stride_cn,
        BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr,
        OUTPUT_F16: tl.constexpr
):
    # Create block pointers for A, B, and C matrices
    a_block_ptr = tl.make_block_ptr(base=a_ptr, shape=(M, K), strides=(stride_am, stride_ak), offsets=(0, 0),
                                    block_shape=(BLOCK_M, BLOCK_K), order=(1, 0))
    b_block_ptr = tl.make_block_ptr(base=b_ptr, shape=(K, N), strides=(stride_bk, stride_bn), offsets=(0, 0),
                                    block_shape=(BLOCK_K, BLOCK_N), order=(0, 1))
    c_block_ptr = tl.make_block_ptr(base=c_ptr, shape=(M, N), strides=(stride_cm, stride_cn), offsets=(0, 0),
                                    block_shape=(BLOCK_M, BLOCK_N), order=(1, 0))
    # Load A and B blocks
    a = tl.load(a_block_ptr)
    b = tl.load(b_block_ptr)

    # Compute matrix product
    c = tl.dot(a, b)
    # Optionally convert the result to float16
    if OUTPUT_F16:
        c = c.to(tl.float16)

    # Store the result
    tl.store(c_block_ptr, c)


def warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A, TRANS_B, OUTPUT_F16):
    # Prepare input matrices
    if (TRANS_A):
        a = torch.randn((K, M), device='cuda', dtype=torch.float16).T
    else:
        a = torch.randn((M, K), device='cuda', dtype=torch.float16)
    if (TRANS_B):
        b = torch.randn((N, K), device='cuda', dtype=torch.float16).T
    else:
        b = torch.randn((K, N), device='cuda', dtype=torch.float16)

    # Prepare output matrix
    c = torch.empty((M, N), device=a.device, dtype=torch.float32)
    if OUTPUT_F16:
        c = torch.empty((M, N), device=a.device, dtype=torch.float16)

    # Execute Triton kernel
    matmul_tma_load_store[(1, 1)](
        a_ptr=a, b_ptr=b, c_ptr=c,
        M=M, N=N, K=K,
        stride_am=a.stride(0), stride_ak=a.stride(1),
        stride_bk=b.stride(0), stride_bn=b.stride(1),
        stride_cm=c.stride(0), stride_cn=c.stride(1),
        BLOCK_M=M, BLOCK_N=N, BLOCK_K=K,
        num_warps=NUM_WARPS, num_ctas=NUM_CTAS,
        OUTPUT_F16=OUTPUT_F16)
    return c
    



##################################################################################################################################################


import torch

def test_all_branches():
    M, N, K = 128, 128, 128
    NUM_CTAS = 1
    NUM_WARPS = 4
    
    results = {}

    # Test case 1: No transposition, output in float32
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=False, TRANS_B=False, OUTPUT_F16=False)
    results["test_case_1"] = out

    # Test case 2: Transpose A, no transpose B, output in float32
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=True, TRANS_B=False, OUTPUT_F16=False)
    results["test_case_2"] = out

    # Test case 3: No transpose A, transpose B, output in float32
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=False, TRANS_B=True, OUTPUT_F16=False)
    results["test_case_3"] = out

    # Test case 4: Transpose A, transpose B, output in float32
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=True, TRANS_B=True, OUTPUT_F16=False)
    results["test_case_4"] = out

    # Test case 5: No transposition, output in float16
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=False, TRANS_B=False, OUTPUT_F16=True)
    results["test_case_5"] = out

    # Test case 6: Transpose A, no transpose B, output in float16
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=True, TRANS_B=False, OUTPUT_F16=True)
    results["test_case_6"] = out

    # Test case 7: No transpose A, transpose B, output in float16
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=False, TRANS_B=True, OUTPUT_F16=True)
    results["test_case_7"] = out

    # Test case 8: Transpose A, transpose B, output in float16
    out = warpper_tma_load_store(M, N, K, NUM_CTAS, NUM_WARPS, TRANS_A=True, TRANS_B=True, OUTPUT_F16=True)
    results["test_case_8"] = out

    return results

# Execute the test function and store the result in result_gold
result_gold = test_all_branches()

# print(result_gold)


# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.
######################################## Imports ######################################## 
import numpy as np
import pytest
import torch
from numpy.random import RandomState

import triton
import triton.language as tl

######################################## Imports ######################################## 


@triton.jit
def chained_matmul_kernel(A,  # shape: (m, k)
                            B,  # shape: (n, k)
                            C,  # shape: (n, k)
                            out,  # shape: (m, k)
                            m, n, k: tl.constexpr,  #
                            block_m: tl.constexpr, block_n: tl.constexpr, block_k: tl.constexpr):

    tl.static_assert(block_k == k, f"expected block_k == k but got {block_k} != {k}")

    block_ix = tl.program_id(0)
    a_tile = (block_ix * block_m + tl.arange(0, block_m))[:, None] * block_k \
        + tl.arange(0, block_k)[None, :]

    a = tl.load(A + a_tile, mask=a_tile < m * k, other=0.0)

    acc = tl.zeros([block_m, block_k], dtype=tl.float32)

    for loop_block_start in range(0, n, block_n):
        bc_tile = (loop_block_start + tl.arange(0, block_n))[:, None] * block_k \
            + tl.arange(0, block_k)[None, :]
        b = tl.load(B + bc_tile, mask=bc_tile < n * k, other=0.0)

        intermediate = tl.dot(a, tl.trans(b))
        intermediate_mask = ((loop_block_start + tl.arange(0, block_n)) < n)[None, :] \
            * (tl.arange(0, block_m) < m)[:, None]

        intermediate = tl.where(intermediate_mask, intermediate, 0.0)

        c = tl.load(C + bc_tile, mask=bc_tile < n * k)

        acc += tl.dot(intermediate.to(A.dtype.element_ty), c)

    tl.store(out + a_tile, acc.to(A.dtype.element_ty), mask=a_tile < m * k)


##################################################################################################################################################  

import numpy as np
import random
import torch 
import os
import pytest
from numpy.random import RandomState
from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict

result_gold = {}

######################################## HELPERS for Eval ######################################## 
def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

######################################## HELPERS for Eval ######################################## 


def chained_matmul_reference(a, b, c):
    intermediate = torch.einsum('MK,NK->MN', a, b)
    return torch.einsum('MN,NK->MK', intermediate, c)


def test_chained_matmul(request, device='cuda'):
    # Regression test for issue #1601
    set_seed()


    m, n, k = 32, 64, 128
    block_m, block_n, block_k = 16, 32, k

    grid = (triton.cdiv(m, block_m), )
    a = torch.randint(low=0, high=2, size=(m, k), dtype=torch.float16, device=device)
    b = torch.randint(low=0, high=2, size=(n, k), dtype=torch.float16, device=device)
    c = torch.randint_like(b, low=0, high=2)
    triton_result = torch.zeros_like(a)

    torch_result = chained_matmul_reference(a, b, c)
    chained_matmul_kernel[grid](
        a, b, c, triton_result, m, n, k,  #
        block_m=block_m, block_n=block_n, block_k=block_k)

    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = triton_result.clone().detach().cpu()
    ################################################################### 
    assert (torch_result == triton_result).all()

# --- Python wrapper for the kernel ---
def chained_matmul_triton_wrapper(A_in, B_in, C_in, out_buffer, block_m, block_n):
    m, k_a = A_in.shape
    n, k_b = B_in.shape
    _n_c, k_c = C_in.shape # C is also (n,k)
    assert k_a == k_b and k_a == k_c, "K dimensions must match for A, B, C"
    
    # Kernel's k is a constexpr fixed to k_a (full K dimension)
    # Kernel's block_k is also fixed to k_a
    
    grid = (triton.cdiv(m, block_m),) # Grid is 1D, iterating over M blocks
    
    chained_matmul_kernel[grid](
        A_in, B_in, C_in, out_buffer,
        m, n, k_a, # runtime m, n, k
        block_m=block_m, block_n=block_n, block_k=k_a # constexpr block_m, block_n, block_k=k
        # num_warps not in kernel signature, would be for autotune
    )
    return out_buffer

# --- Define TFLOPS and GB/s calculators ---
def calculate_chained_matmul_tflops(params: dict, ms: float) -> float:
    m, n, k = params['M'], params['N'], params['K']
    # Op1: Intermediate(M,N) = A(M,K) @ B.T(K,N) -> 2 * M * N * K FLOPs
    # Op2: Out(M,K) = Intermediate(M,N) @ C(N,K) -> 2 * M * K * N FLOPs (Note: C is (N,K))
    # Total FLOPs = 4 * M * N * K
    flops = 2 * m * n * k + 2 * m * k * n 
    tflops = flops / (ms / 1000) / 1e12
    return tflops

def calculate_chained_matmul_gbps(params: dict, ms: float) -> float:
    m, n, k = params['M'], params['N'], params['K']
    dtype_str = params.get('dtype_str', 'fp16')
    current_dtype = torch.float16
    if dtype_str == 'fp32': current_dtype = torch.float32
    elif dtype_str == 'bf16': current_dtype = torch.bfloat16
    element_size = torch.tensor([], dtype=current_dtype).element_size()
    bytes_a = m * k * element_size
    bytes_b = n * k * element_size
    bytes_c_read = n * k * element_size
    bytes_out_write = m * k * element_size
    total_bytes = bytes_a + bytes_b + bytes_c_read + bytes_out_write
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

OP_NAME_FOR_BENCHMARK = "chained_matmul_triton_perf"

# --- REVISED Pytest parametrize for performance testing ---
# K must be small enough for block_k = K to fit in shared memory.
# Constraint: K * (block_m + block_n) <= 32768 (for fp16, 64KB shared mem limit)

CM_PERF_TEST_CONFIGS = []
# Shared memory constraint (fp16): block_m*K + 2*K*block_n + block_m*block_n <= 32768

# K = 64
k_val = 64
for M_val in [64, 128, 256]:
    for N_val in [64, 128, 256]: # N is the total dimension, looped over by block_n
        for bm_val in [16, 32, 64]:
            if M_val % bm_val == 0:
                for bn_val in [16, 32]: # Keep block_n small
                    if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768:
                        CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})
                # Try bn_val = 64 if bm_val is small
                if bm_val <= 32:
                    bn_val = 64
                    if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768:
                        CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})


# K = 128
k_val = 128
for M_val in [64, 128, 256]:
    for N_val in [32, 64, 128]: 
        for bm_val in [16, 32]: # Keep block_m smaller for larger K
            if M_val % bm_val == 0:
                for bn_val in [16, 32]: # Keep block_n small
                    if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768:
                        CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})
        # Try bm_val = 64 if bn_val is very small
        if M_val % 64 == 0:
            bm_val = 64
            bn_val = 16 
            if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768:
                 CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})


# K = 256 (Requires even smaller block_m, block_n)
k_val = 256
for M_val in [32, 64, 128]: 
    for N_val in [16, 32, 64]:  
        for bm_val in [16, 32]: 
            if M_val % bm_val == 0:
                for bn_val in [16]: # block_n must be very small
                    if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768:
                        CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})
                # Try bn_val = 32 if bm_val is 16
                if bm_val == 16:
                    bn_val = 32
                    if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768: # 16*256 + 2*256*32 + 16*32 = 4096 + 16384 + 512 = 20992 (OK)
                        CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})


# K = 512 (Extremely restrictive for this kernel design) - Likely only 16x16 blocks will work
k_val = 512
for M_val in [16, 32, 64]: 
    for N_val in [16, 32]: 
        for bm_val in [16]: 
            if M_val % bm_val == 0:
                for bn_val in [16]: 
                    # 16*512 + 2*512*16 + 16*16 = 8192 + 16384 + 256 = 24832 (OK)
                    if (bm_val*k_val + 2*k_val*bn_val + bm_val*bn_val) <= 32768:
                        CM_PERF_TEST_CONFIGS.append({'M': M_val, 'N': N_val, 'K': k_val, 'bm': bm_val, 'bn': bn_val})


# Remove duplicates
unique_configs = []
seen_configs_str = set()
for cfg in CM_PERF_TEST_CONFIGS:
    # Create a string representation for checking uniqueness easily
    cfg_str = f"M{cfg['M']}_N{cfg['N']}_K{cfg['K']}_bm{cfg['bm']}_bn{cfg['bn']}"
    if cfg_str not in seen_configs_str:
        unique_configs.append(cfg)
        seen_configs_str.add(cfg_str)
CM_PERF_TEST_CONFIGS = unique_configs

print(f"Generated {len(CM_PERF_TEST_CONFIGS)} unique performance test configurations for chained_matmul.")

# Only fp16 for now to reduce matrix and focus on shared mem issue
CM_DTYPES_FOR_PERF = ['fp16']
# CM_DTYPES_FOR_PERF = ['fp16', 'fp32'] # Add fp32 later if fp16 works


@pytest.mark.parametrize("test_params_dict", CM_PERF_TEST_CONFIGS)
@pytest.mark.parametrize("dtype_str", CM_DTYPES_FOR_PERF)
def test_performance(test_params_dict, dtype_str, request):
    # ... (rest of the test_chained_matmul_performance function remains the same as in the previous response)
    set_seed()
    m = test_params_dict['M']
    n = test_params_dict['N']
    k = test_params_dict['K']
    block_m = test_params_dict['bm']
    block_n = test_params_dict['bn']
    
    block_k_const = k 

    # This skip logic might be redundant if CM_PERF_TEST_CONFIGS is well-filtered
    # element_size_bytes = 4 if dtype_str == 'fp32' else 2
    # shared_mem_limit_elements = 65536 // element_size_bytes
    # estimated_elements_needed = block_m * k + 2 * k * block_n + block_m * block_n
    # if estimated_elements_needed > shared_mem_limit_elements:
    #     pytest.skip(f"Skipping M{m}N{n}K{k} bm{block_m}bn{block_n} dtype {dtype_str} due to estimated shared memory: {estimated_elements_needed*element_size_bytes} vs {65536}")


    if dtype_str == 'fp32': current_dtype = torch.float32
    elif dtype_str == 'bf16': current_dtype = torch.bfloat16
    else: current_dtype = torch.float16

    a = torch.randn((m, k), dtype=current_dtype, device='cuda')
    b = torch.randn((n, k), dtype=current_dtype, device='cuda') 
    c_mat = torch.randn((n, k), dtype=current_dtype, device='cuda')
    triton_result_buffer = torch.empty((m, k), dtype=current_dtype, device='cuda')

    op_lambda = lambda: chained_matmul_triton_wrapper(
        a, b, c_mat, triton_result_buffer,
        block_m, block_n
    )

    bench_config = do_bench_config(warm_up=25, repetition=100)
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)

    current_params_for_logs_and_calc = {
        "M": m, "N": n, "K": k,
        "block_m": block_m, "block_n": block_n, "block_k": block_k_const,
        "dtype_str": dtype_str
    }

    benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                              gbps_calculator=calculate_chained_matmul_gbps,
                              tflops_calculator=calculate_chained_matmul_tflops)
######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all y_triton results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} y_triton tensors to {OUTPUT_FILENAME}.")  

def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")

######################################## HELPERS for Eval ########################################


# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
        triton.Config({}, num_warps=16),
        # triton.Config({}, num_warps=32),
    ],
    key=["N", "HAS_RESIDUAL", "STORE_RESIDUAL_OUT", "IS_RMS_NORM", "HAS_BIAS"],
)
@triton.heuristics({"HAS_X1": lambda args: args["X1"] is not None})
@triton.heuristics({"HAS_W1": lambda args: args["W1"] is not None})
@triton.heuristics({"HAS_B1": lambda args: args["B1"] is not None})
@triton.jit
def _layer_norm_fwd_1pass_kernel(
    X, Y, W, B, RESIDUAL, X1, W1, B1, Y1, RESIDUAL_OUT, ROWSCALE, SEEDS, DROPOUT_MASK, Mean, Rstd,
    stride_x_row, stride_y_row, stride_res_row, stride_res_out_row, stride_x1_row, stride_y1_row,
    M, N, eps, dropout_p, IS_RMS_NORM: tl.constexpr, BLOCK_N: tl.constexpr, HAS_RESIDUAL: tl.constexpr,
    STORE_RESIDUAL_OUT: tl.constexpr, HAS_BIAS: tl.constexpr, HAS_DROPOUT: tl.constexpr,
    STORE_DROPOUT_MASK: tl.constexpr, HAS_ROWSCALE: tl.constexpr, HAS_X1: tl.constexpr,
    HAS_W1: tl.constexpr, HAS_B1: tl.constexpr,
):
    row = tl.program_id(0)
    X += row * stride_x_row
    Y += row * stride_y_row
    if HAS_RESIDUAL:
        RESIDUAL += row * stride_res_row
    if STORE_RESIDUAL_OUT:
        RESIDUAL_OUT += row * stride_res_out_row
    if HAS_X1:
        X1 += row * stride_x1_row
    if HAS_W1:
        Y1 += row * stride_y1_row
    cols = tl.arange(0, BLOCK_N)
    x = tl.load(X + cols, mask=cols < N, other=0.0).to(tl.float32)
    if HAS_ROWSCALE:
        rowscale = tl.load(ROWSCALE + row).to(tl.float32)
        x *= rowscale
    if HAS_DROPOUT:
        keep_mask = tl.rand(tl.load(SEEDS + row).to(tl.uint32), cols, n_rounds=7) > dropout_p
        x = tl.where(keep_mask, x / (1.0 - dropout_p), 0.0)
        if STORE_DROPOUT_MASK:
            tl.store(DROPOUT_MASK + row * N + cols, keep_mask, mask=cols < N)
    if HAS_X1:
        x1 = tl.load(X1 + cols, mask=cols < N, other=0.0).to(tl.float32)
        if HAS_ROWSCALE:
            rowscale = tl.load(ROWSCALE + M + row).to(tl.float32)
            x1 *= rowscale
        if HAS_DROPOUT:
            keep_mask = (
                tl.rand(tl.load(SEEDS + M + row).to(tl.uint32), cols, n_rounds=7) > dropout_p
            )
            x1 = tl.where(keep_mask, x1 / (1.0 - dropout_p), 0.0)
            if STORE_DROPOUT_MASK:
                tl.store(DROPOUT_MASK + (M + row) * N + cols, keep_mask, mask=cols < N)
        x += x1
    if HAS_RESIDUAL:
        residual = tl.load(RESIDUAL + cols, mask=cols < N, other=0.0).to(tl.float32)
        x += residual
    if STORE_RESIDUAL_OUT:
        tl.store(RESIDUAL_OUT + cols, x, mask=cols < N)
    if not IS_RMS_NORM:
        mean = tl.sum(x, axis=0) / N
        tl.store(Mean + row, mean)
        xbar = tl.where(cols < N, x - mean, 0.0)
        var = tl.sum(xbar * xbar, axis=0) / N
    else:
        xbar = tl.where(cols < N, x, 0.0)
        var = tl.sum(xbar * xbar, axis=0) / N
    rstd = 1 / tl.sqrt(var + eps)
    tl.store(Rstd + row, rstd)
    mask = cols < N
    w = tl.load(W + cols, mask=mask).to(tl.float32)
    if HAS_BIAS:
        b = tl.load(B + cols, mask=mask).to(tl.float32)
    x_hat = (x - mean) * rstd if not IS_RMS_NORM else x * rstd
    y = x_hat * w + b if HAS_BIAS else x_hat * w
    tl.store(Y + cols, y, mask=mask)
    if HAS_W1:
        w1 = tl.load(W1 + cols, mask=mask).to(tl.float32)
        if HAS_B1:
            b1 = tl.load(B1 + cols, mask=mask).to(tl.float32)
        y1 = x_hat * w1 + b1 if HAS_B1 else x_hat * w1
        tl.store(Y1 + cols, y1, mask=mask)

def _layer_norm_fwd(
    x, weight, bias, eps, residual=None, x1=None, weight1=None, bias1=None, dropout_p=0.0,
    rowscale=None, out_dtype=None, residual_dtype=None, is_rms_norm=False, return_dropout_mask=False,
):
    if residual is not None:
        residual_dtype = residual.dtype
    M, N = x.shape
    assert x.stride(-1) == 1
    if residual is not None:
        assert residual.stride(-1) == 1
        assert residual.shape == (M, N)
    assert weight.shape == (N,)
    assert weight.stride(-1) == 1
    if bias is not None:
        assert bias.stride(-1) == 1
        assert bias.shape == (N,)
    if x1 is not None:
        assert x1.shape == x.shape
        assert rowscale is None
        assert x1.stride(-1) == 1
    if weight1 is not None:
        assert weight1.shape == (N,)
        assert weight1.stride(-1) == 1
    if bias1 is not None:
        assert bias1.shape == (N,)
        assert bias1.stride(-1) == 1
    if rowscale is not None:
        assert rowscale.is_contiguous()
        assert rowscale.shape == (M,)
    y = torch.empty_like(x, dtype=x.dtype if out_dtype is None else out_dtype)
    assert y.stride(-1) == 1
    if weight1 is not None:
        y1 = torch.empty_like(y)
        assert y1.stride(-1) == 1
    else:
        y1 = None
    if (
        residual is not None
        or (residual_dtype is not None and residual_dtype != x.dtype)
        or dropout_p > 0.0
        or rowscale is not None
        or x1 is not None
    ):
        residual_out = torch.empty(
            M, N, device=x.device, dtype=residual_dtype if residual_dtype is not None else x.dtype
        )
        assert residual_out.stride(-1) == 1
    else:
        residual_out = None
    mean = torch.empty((M,), dtype=torch.float32, device=x.device) if not is_rms_norm else None
    rstd = torch.empty((M,), dtype=torch.float32, device=x.device)
    if dropout_p > 0.0:
        seeds = torch.randint(
            2**32, (M if x1 is None else 2 * M,), device=x.device, dtype=torch.int64
        )
    else:
        seeds = None
    if return_dropout_mask and dropout_p > 0.0:
        dropout_mask = torch.empty(M if x1 is None else 2 * M, N, device=x.device, dtype=torch.bool)
    else:
        dropout_mask = None
    MAX_FUSED_SIZE = 65536 // x.element_size()
    BLOCK_N = min(MAX_FUSED_SIZE, triton.next_power_of_2(N))
    if N > BLOCK_N:
        raise RuntimeError("This layer norm doesn't support feature dim >= 64KB.")
    with torch.cuda.device(x.device.index):
        _layer_norm_fwd_1pass_kernel[(M,)](
            x, y, weight, bias, residual, x1, weight1, bias1, y1, residual_out, rowscale, seeds,
            dropout_mask, mean, rstd, x.stride(0), y.stride(0), residual.stride(0) if residual is not None else 0,
            residual_out.stride(0) if residual_out is not None else 0, x1.stride(0) if x1 is not None else 0,
            y1.stride(0) if y1 is not None else 0, M, N, eps, dropout_p, is_rms_norm, BLOCK_N,
            residual is not None, residual_out is not None, bias is not None, dropout_p > 0.0,
            dropout_mask is not None, rowscale is not None,
        )
    if dropout_mask is not None and x1 is not None:
        dropout_mask, dropout_mask1 = dropout_mask.tensor_split(2, dim=0)
    else:
        dropout_mask1 = None
    return (
        y, y1, mean, rstd, residual_out if residual_out is not None else x, seeds, dropout_mask, dropout_mask1,
    )




##################################################################################################################################################


import torch

def test_layer_norm_fwd():
    # Define the input parameters
    M, N = 64, 128  # Example dimensions
    eps = 1e-5
    dropout_p = 0.1

    # Create random input tensors
    x = torch.randn(M, N, device='cuda', dtype=torch.float32)
    weight = torch.randn(N, device='cuda', dtype=torch.float32)
    bias = torch.randn(N, device='cuda', dtype=torch.float32)
    residual = torch.randn(M, N, device='cuda', dtype=torch.float32)
    x1 = torch.randn(M, N, device='cuda', dtype=torch.float32)
    weight1 = torch.randn(N, device='cuda', dtype=torch.float32)
    bias1 = torch.randn(N, device='cuda', dtype=torch.float32)
    rowscale = torch.randn(M, device='cuda', dtype=torch.float32)

    results = {}

    # Test case 1: Basic layer norm
    results['test_case_1'] = _layer_norm_fwd(x, weight, bias, eps)

    # Test case 2: Layer norm with residual
    results['test_case_2'] = _layer_norm_fwd(x, weight, bias, eps, residual=residual)

    # Test case 3: Layer norm with additional input tensors
    results['test_case_3'] = _layer_norm_fwd(x, weight, bias, eps, x1=x1, weight1=weight1, bias1=bias1)

    # Test case 4: Layer norm with dropout
    results['test_case_4'] = _layer_norm_fwd(x, weight, bias, eps, dropout_p=dropout_p, return_dropout_mask=True)

    # Test case 5: Layer norm with row scaling
    results['test_case_5'] = _layer_norm_fwd(x, weight, bias, eps, rowscale=rowscale)

    return results

result_gold = test_layer_norm_fwd()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import math
import triton.language as tl

@triton.jit(do_not_specialize=["eps"])
def rms_norm_kernel(
    Y,  # pointer to the output
    X,  # pointer to the input
    W,  # pointer to the weights
    y_stride_r,
    y_stride_c,
    x_stride_r,  # how much to increase the pointer when moving by 1 row
    x_stride_c,  # how much to increase the pointer when moving by 1 col
    N,  # number of columns in X
    eps,  # epsilon to avoid division by zero
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(0)
    Y += pid * y_stride_r
    X += pid * x_stride_r

    mask = tl.arange(0, BLOCK_SIZE) < N
    cols = tl.arange(0, BLOCK_SIZE)
    x = tl.load(X + cols * x_stride_c, mask, other=0.0).to(tl.float32)

    var = tl.sum(x * x, axis=0) / N
    rrms = 1 / tl.sqrt(var + eps)

    w = tl.load(W + tl.arange(0, BLOCK_SIZE), mask=mask, other=0.0)
    y = (x * rrms).to(Y.dtype.element_ty) * w
    tl.store(Y + cols * y_stride_c, y, mask=mask)


class RmsNorm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, normalized_shape, weight, eps=1e-5):
        dim = x.ndim - len(normalized_shape)
        M = math.prod(x.shape[:dim])
        N = math.prod(normalized_shape)

        BLOCK_SIZE = triton.next_power_of_2(N)
        x = x.contiguous()
        weight = weight.contiguous()
        y = torch.empty_like(x)

        with torch.cuda.device(x.device):
            rms_norm_kernel[M,](y, x, weight, N, 1, N, 1, N, eps, BLOCK_SIZE)
        return y


def rms_norm(x, normalized_shape, weight, eps=1e-5):
    return RmsNorm.apply(x, normalized_shape, weight, eps)



##################################################################################################################################################


def test_rms_norm():
    # Define input parameters
    batch_size = 32
    feature_size = 128
    eps = 1e-5

    # Create random input data and weights
    x = torch.randn(batch_size, feature_size, device='cuda', dtype=torch.float32)
    weight = torch.randn(feature_size, device='cuda', dtype=torch.float32)

    # Triton implementation
    output = rms_norm(x, (feature_size,), weight, eps)

    # Additional test cases to cover all branches
    test_case_1 = rms_norm(x, (feature_size,), weight, eps)
    test_case_2 = rms_norm(x, (feature_size,), weight, eps=1e-6)
    test_case_3 = rms_norm(x, (feature_size,), weight, eps=1e-7)
    test_case_4 = rms_norm(x, (feature_size,), weight, eps=1e-8)

    return {
        "test_case_1": test_case_1,
        "test_case_2": test_case_2,
        "test_case_3": test_case_3,
        "test_case_4": test_case_4
    }

result_gold = test_rms_norm()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch

@triton.jit
def sin_kernel(
    in_ptr0,
    out_ptr,
    n_elements,
    BLOCK_SIZE: "tl.constexpr",
):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(in_ptr0 + offsets, mask=mask)
    output = tl.sin(x)
    tl.store(out_ptr + offsets, output, mask=mask)

def sin_triton(x, out):
    n_elements = x.numel()
    sin_kernel[(n_elements,)](x, out, n_elements, BLOCK_SIZE=4)



##################################################################################################################################################


def test_sin_triton():
    results = {}
    
    # Test case 1
    x1 = torch.tensor([0.0, 1.0, 2.0, 3.0], device='cuda')
    out1 = torch.empty_like(x1)
    sin_triton(x1, out1)
    results['test_case_1'] = out1

    # Test case 2
    x2 = torch.tensor([4.0, 5.0, 6.0, 7.0], device='cuda')
    out2 = torch.empty_like(x2)
    sin_triton(x2, out2)
    results['test_case_2'] = out2

    # Test case 3
    x3 = torch.tensor([8.0, 9.0, 10.0, 11.0], device='cuda')
    out3 = torch.empty_like(x3)
    sin_triton(x3, out3)
    results['test_case_3'] = out3

    # Test case 4
    x4 = torch.tensor([12.0, 13.0, 14.0, 15.0], device='cuda')
    out4 = torch.empty_like(x4)
    sin_triton(x4, out4)
    results['test_case_4'] = out4

    return results

result_gold = test_sin_triton()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch

import triton
import triton.language as tl


torch.manual_seed(1234)


@triton.jit
def rms_matmul_rbe(
        x_ptr, w_ptr, rms_w_ptr, out_ptr,
        M, N, K,
        stride_x_batch, stride_x_m, stride_x_k,
        stride_w_k, stride_w_n,
        stride_rms_w,
        stride_out_batch, stride_out_m, stride_out_n,
        start_token_position,
        USE_FP8: tl.constexpr,
        RBE_EPILOGUE: tl.constexpr,
        THETA: tl.constexpr,
        EPS: tl.constexpr,
        BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
):
    """
    Prologue: RMS
    Epilogue: nothing or Rotary embeddings
    c = ROBE((rms(a) * rms_w) @ b)
    """
    pid_batch = tl.program_id(axis=0)
    pid = tl.program_id(axis=1)
    pid_m = pid // tl.cdiv(N, BLOCK_SIZE_N)
    pid_n = pid % tl.cdiv(N, BLOCK_SIZE_N)

    offs_m = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_n = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    x_ptrs = x_ptr + (pid_batch * stride_x_batch + offs_m[:, None] * stride_x_m + offs_k[None, :] * stride_x_k)
    w_ptrs = w_ptr + (offs_k[:, None] * stride_w_k + offs_n[None, :] * stride_w_n)
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    rms_w_ptrs = rms_w_ptr + tl.arange(0, BLOCK_SIZE_K)[None, :] * stride_rms_w
    x_sum = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_K), dtype=tl.float32)
    for _ in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        x = tl.load(x_ptrs)
        x_sum += tl.extra.hip.libdevice.pow(x.to(tl.float32), 2)
        rms_w = tl.load(rms_w_ptrs)  # TODO add an assert that rms_w is a multiple of BLOCK SIZE K
        if USE_FP8:
            rms_w = rms_w.to(tl.float8e5, bitcast=True)
            rms_w = rms_w.to(tl.float16)
        x = x * rms_w
        w = tl.load(w_ptrs)  # TODO add an assert that w is a multiple of BLOCK SIZE K
        if USE_FP8:
            w = w.to(tl.float8e5, bitcast=True)
            w = w.to(tl.float32)
            w = w.to(tl.float16)
        accumulator += tl.dot(x, w)
        x_ptrs += BLOCK_SIZE_K * stride_x_k
        w_ptrs += BLOCK_SIZE_K * stride_w_k
        rms_w_ptrs += BLOCK_SIZE_K * stride_rms_w
    x_mean = tl.sum(x_sum, axis=1) / K + EPS
    x_norm = tl.math.rsqrt(x_mean)
    accumulator = accumulator * x_norm[:, None]

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    out_ptrs = out_ptr + (
                pid_batch * stride_out_batch + offs_m[:, None] * stride_out_m + offs_n[None, :] * stride_out_n)
    out_mask = (offs_m[:, None] < M) & (offs_n[None, :] < N)

    tl.store(out_ptrs, accumulator, mask=out_mask)


@triton.jit
def rms_matmul_rbe_qkv(x_ptr,
                       q_weight_ptr, k_weight_ptr, v_weight_ptr,
                       rms_w_ptr,
                       q_ptr, k_ptr, v_ptr,
                       M, N, K,
                       stride_x_batch, stride_x_m, stride_x_k,
                       stride_q_w_k, stride_q_w_n,
                       stride_k_w_k, stride_k_w_n,
                       stride_v_w_k, stride_v_w_n,
                       stride_rms_w,
                       stride_q_batch, stride_q_m, stride_q_n,
                       stride_k_batch, stride_k_m, stride_k_n,
                       stride_v_batch, stride_v_m, stride_v_n,
                       start_token_position,
                       USE_FP8: tl.constexpr,
                       THETA: tl.constexpr,
                       EPS: tl.constexpr,
                       BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr):
    # q
    rms_matmul_rbe(
        x_ptr=x_ptr,
        w_ptr=q_weight_ptr, rms_w_ptr=rms_w_ptr, out_ptr=q_ptr,
        M=M, N=N, K=K,
        stride_x_batch=stride_x_batch, stride_x_m=stride_x_m, stride_x_k=stride_x_k,
        stride_w_k=stride_q_w_k, stride_w_n=stride_q_w_n,
        stride_rms_w=stride_rms_w,
        stride_out_batch=stride_q_batch, stride_out_m=stride_q_m, stride_out_n=stride_q_n,
        start_token_position=start_token_position,
        USE_FP8=USE_FP8,
        RBE_EPILOGUE=True,
        THETA=THETA,
        EPS=EPS,
        BLOCK_SIZE_M=BLOCK_SIZE_M, BLOCK_SIZE_N=BLOCK_SIZE_N, BLOCK_SIZE_K=BLOCK_SIZE_K,
    )
    # k
    rms_matmul_rbe(
        x_ptr=x_ptr,
        w_ptr=k_weight_ptr, rms_w_ptr=rms_w_ptr, out_ptr=k_ptr,
        M=M, N=N, K=K,
        stride_x_batch=stride_x_batch, stride_x_m=stride_x_m, stride_x_k=stride_x_k,
        stride_w_k=stride_k_w_k, stride_w_n=stride_k_w_n,
        stride_rms_w=stride_rms_w,
        stride_out_batch=stride_k_batch, stride_out_m=stride_k_m, stride_out_n=stride_k_n,
        start_token_position=start_token_position,
        USE_FP8=USE_FP8,
        RBE_EPILOGUE=True,
        THETA=THETA,
        EPS=EPS,
        BLOCK_SIZE_M=BLOCK_SIZE_M, BLOCK_SIZE_N=BLOCK_SIZE_N, BLOCK_SIZE_K=BLOCK_SIZE_K,
    )
    # v
    rms_matmul_rbe(
        x_ptr=x_ptr,
        w_ptr=v_weight_ptr, rms_w_ptr=rms_w_ptr, out_ptr=v_ptr,
        M=M, N=N, K=K,
        stride_x_batch=stride_x_batch, stride_x_m=stride_x_m, stride_x_k=stride_x_k,
        stride_w_k=stride_v_w_k, stride_w_n=stride_v_w_n,
        stride_rms_w=stride_rms_w,
        stride_out_batch=stride_v_batch, stride_out_m=stride_v_m, stride_out_n=stride_v_n,
        start_token_position=start_token_position,
        USE_FP8=USE_FP8,
        RBE_EPILOGUE=False,
        THETA=THETA,
        EPS=EPS,
        BLOCK_SIZE_M=BLOCK_SIZE_M, BLOCK_SIZE_N=BLOCK_SIZE_N, BLOCK_SIZE_K=BLOCK_SIZE_K,
    )


def rms_matmul_rbe_qkv_wrapper(x: torch.Tensor,
                               start_pos: int,
                               q_weight: torch.Tensor, k_weight: torch.Tensor, v_weight: torch.Tensor,
                               rms_w: torch.Tensor,
                               n_heads: int, head_dim: int,
                               k: torch.Tensor,
                               v: torch.Tensor,
                               eps: float = 1e-6, theta=10000.):
    assert q_weight.shape == k_weight.shape == v_weight.shape
    assert q_weight.dtype == k_weight.dtype == v_weight.dtype == rms_w.dtype
    assert q_weight.dtype in [torch.float16, torch.int8]
    batch, M, K = x.shape

    assert K == rms_w.shape[0]

    q_weight_t = q_weight.t()
    k_weight_t = k_weight.t()
    v_weight_t = v_weight.t()
    K_W, N = q_weight_t.shape
    assert K == K_W
    q = torch.empty((batch, M, N), dtype=torch.float16, device=q_weight_t.device)

    k = k.view((batch, M, N))
    v = v.view((batch, M, N))
    assert k.dtype == k_weight.dtype
    assert v.dtype == v_weight.dtype

    q_ptr = triton.reinterpret(q, tl.float16)
    k_ptr = triton.reinterpret(k, tl.float8e5 if k.dtype == torch.int8 else tl.float16)
    v_ptr = triton.reinterpret(v, tl.float8e5 if v.dtype == torch.int8 else tl.float16)

    grid = lambda META: (
    batch, triton.cdiv(META["M"], META["BLOCK_SIZE_M"]) * triton.cdiv(META["N"], META["BLOCK_SIZE_N"]))

    rms_matmul_rbe_qkv[grid](
        x_ptr=x,
        q_weight_ptr=q_weight_t, k_weight_ptr=k_weight_t, v_weight_ptr=v_weight_t,
        rms_w_ptr=rms_w,
        q_ptr=q_ptr, k_ptr=k_ptr, v_ptr=v_ptr,
        M=M, N=N, K=K,
        stride_x_batch=x.stride(0), stride_x_m=x.stride(1), stride_x_k=x.stride(2),
        stride_q_w_k=q_weight_t.stride(0), stride_q_w_n=q_weight_t.stride(1),
        stride_k_w_k=k_weight_t.stride(0), stride_k_w_n=k_weight_t.stride(1),
        stride_v_w_k=v_weight_t.stride(0), stride_v_w_n=v_weight_t.stride(1),
        stride_rms_w=rms_w.stride(0),
        stride_q_batch=q.stride(0), stride_q_m=q.stride(1), stride_q_n=q.stride(2),
        stride_k_batch=k.stride(0), stride_k_m=k.stride(1), stride_k_n=k.stride(2),
        stride_v_batch=v.stride(0), stride_v_m=v.stride(1), stride_v_n=v.stride(2),
        start_token_position=start_pos,
        USE_FP8=q_weight.dtype == torch.int8,
        THETA=theta,
        EPS=eps,
        BLOCK_SIZE_M=16, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64,
        num_stages=4, num_warps=4
    )
    q = q.view(batch, M, n_heads, head_dim)
    k = k.view(batch, M, n_heads, head_dim)
    v = v.view(batch, M, n_heads, head_dim)
    return q, k, v




##################################################################################################################################################


# Test for rms_matmul_rbe_qkv_wrapper
def test_rms_matmul_rbe_qkv():
    results = {}
    
    # Test case 1
    batch, seq_len, heads, dim = [1, 16, 32, 128]
    embeddings_load = torch.randn([batch, seq_len, heads * dim], dtype=torch.float16, device="cuda")
    rms_weights = torch.randn([heads * dim], dtype=torch.float16, device="cuda") * 0.2
    q_weights_load = torch.randn([heads * dim, heads * dim], dtype=torch.float16, device="cuda") * 0.2
    k = torch.empty((embeddings_load.shape[0], embeddings_load.shape[1], q_weights_load.shape[-1]),
                    dtype=q_weights_load.dtype, device=q_weights_load.device)
    v = torch.empty_like(k)
    q, k, v = rms_matmul_rbe_qkv_wrapper(x=embeddings_load, start_pos=0,
                                         q_weight=q_weights_load, k_weight=q_weights_load,
                                         v_weight=q_weights_load, rms_w=rms_weights,
                                         k=k, v=v,
                                         n_heads=32,
                                         head_dim=128)
    results['test_case_1'] = (q.shape, k.shape, v.shape)
    
    # Test case 2: Different dimensions
    batch, seq_len, heads, dim = [2, 32, 16, 64]
    embeddings_load = torch.randn([batch, seq_len, heads * dim], dtype=torch.float16, device="cuda")
    rms_weights = torch.randn([heads * dim], dtype=torch.float16, device="cuda") * 0.2
    q_weights_load = torch.randn([heads * dim, heads * dim], dtype=torch.float16, device="cuda") * 0.2
    k = torch.empty((embeddings_load.shape[0], embeddings_load.shape[1], q_weights_load.shape[-1]),
                    dtype=q_weights_load.dtype, device=q_weights_load.device)
    v = torch.empty_like(k)
    q, k, v = rms_matmul_rbe_qkv_wrapper(x=embeddings_load, start_pos=0,
                                         q_weight=q_weights_load, k_weight=q_weights_load,
                                         v_weight=q_weights_load, rms_w=rms_weights,
                                         k=k, v=v,
                                         n_heads=16,
                                         head_dim=64)
    results['test_case_2'] = (q.shape, k.shape, v.shape)
    
    # Test case 3: Different start position
    batch, seq_len, heads, dim = [1, 16, 32, 128]
    embeddings_load = torch.randn([batch, seq_len, heads * dim], dtype=torch.float16, device="cuda")
    rms_weights = torch.randn([heads * dim], dtype=torch.float16, device="cuda") * 0.2
    q_weights_load = torch.randn([heads * dim, heads * dim], dtype=torch.float16, device="cuda") * 0.2
    k = torch.empty((embeddings_load.shape[0], embeddings_load.shape[1], q_weights_load.shape[-1]),
                    dtype=q_weights_load.dtype, device=q_weights_load.device)
    v = torch.empty_like(k)
    q, k, v = rms_matmul_rbe_qkv_wrapper(x=embeddings_load, start_pos=5,
                                         q_weight=q_weights_load, k_weight=q_weights_load,
                                         v_weight=q_weights_load, rms_w=rms_weights,
                                         k=k, v=v,
                                         n_heads=32,
                                         head_dim=128)
    results['test_case_3'] = (q.shape, k.shape, v.shape)
    
    # Test case 4: Different weights
    batch, seq_len, heads, dim = [1, 16, 32, 128]
    embeddings_load = torch.randn([batch, seq_len, heads * dim], dtype=torch.float16, device="cuda")
    rms_weights = torch.randn([heads * dim], dtype=torch.float16, device="cuda") * 0.2
    q_weights_load = torch.randn([heads * dim, heads * dim], dtype=torch.float16, device="cuda") * 0.1
    k = torch.empty((embeddings_load.shape[0], embeddings_load.shape[1], q_weights_load.shape[-1]),
                    dtype=q_weights_load.dtype, device=q_weights_load.device)
    v = torch.empty_like(k)
    q, k, v = rms_matmul_rbe_qkv_wrapper(x=embeddings_load, start_pos=0,
                                         q_weight=q_weights_load, k_weight=q_weights_load,
                                         v_weight=q_weights_load, rms_w=rms_weights,
                                         k=k, v=v,
                                         n_heads=32,
                                         head_dim=128)
    results['test_case_4'] = (q.shape, k.shape, v.shape)
    
    return results

result_gold = test_rms_matmul_rbe_qkv()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.chunk_gate_recurrence import chunk_gate_recurrent
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('chunk_gate_recurrence', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 11):  # Adjust the range as needed for your testing
            B = 4
            H = 4
            N = 2 ** i
            D_k = 128
            D_v = 64
            kv = torch.rand(B, H, N, D_k, D_v, dtype=torch.float16)
            cross_decay = torch.rand(B, H, N, dtype=torch.float16)
            self.input_tensors.append((kv, cross_decay))

    def to_cuda(self, input_tensor):
        kv, cross_decay = input_tensor
        return kv.cuda(), cross_decay.cuda()

    def call_op(self, input_tensor):
        kv, cross_decay = input_tensor
        return chunk_gate_recurrent(kv, cross_decay)

    def get_gbps(self, input_tensor, runtime):
        kv, _ = input_tensor
        total_bytes = kv.numel() * kv.element_size() * 2  # Read and write
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        kv, _ = input_tensor
        # Assuming each element involves a multiply and an add
        FLOPS = 2 * kv.numel()
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import math
import torch
import triton
import triton.language as tl

_BLOCK_N = 64
_BLOCK_M = 64

@triton.heuristics(
    {
        "IS_EVEN_M": lambda args: args["N_CTX"] % args["BLOCK_M"] == 0,
        "IS_EVEN_N": lambda args: args["NKV_CTX"] % args["BLOCK_N"] == 0,
    }
)
@triton.jit
def _score_kernel(
    Q, K, M, sm_scale, Out,
    stride_qz, stride_qh, stride_qm, stride_qk,  #
    stride_kz, stride_kh, stride_kn, stride_kk,  #
    stride_oz, stride_oh, stride_on,
    Z, H, H_KV, #
    N_CTX,  #
    ROUND_CTX,
    NKV_CTX,
    sliding_window_offset,
    sliding_window_size,
    SLIDING_WINDOW: tl.constexpr,
    COMPLEMENT_SLIDING_WINDOW: tl.constexpr,
    IS_EVEN_M: tl.constexpr,
    IS_EVEN_N: tl.constexpr,
    BLOCK_M: tl.constexpr,  #
    BLOCK_DMODEL: tl.constexpr,  #
    BLOCK_N: tl.constexpr,  #
):
    start_n = tl.program_id(0)
    off_hz = tl.program_id(1)
    off_z = off_hz // H
    off_h = off_hz % H
    off_hkv = off_h // (H//H_KV)
    q_offset = off_z.to(tl.int64) * stride_qz + off_h.to(tl.int64) * stride_qh
    k_offset = off_z.to(tl.int64) * stride_kz + off_hkv.to(tl.int64) * stride_kh
    m_ptrs = M + off_hz * ROUND_CTX + tl.arange(0, BLOCK_M)
    o = tl.zeros([BLOCK_M], dtype=tl.float32)

    Q_block_ptr = tl.make_block_ptr(
        base=Q + q_offset,
        shape=(N_CTX, BLOCK_DMODEL),
        strides=(stride_qm, stride_qk),
        offsets=(0, 0),
        block_shape=(BLOCK_M, BLOCK_DMODEL),
        order=(1, 0),
    )
    K_block_ptr = tl.make_block_ptr(
        base=K + k_offset,
        shape=(BLOCK_DMODEL, NKV_CTX),
        strides=(stride_kk, stride_kn),
        offsets=(0, start_n * BLOCK_N),
        block_shape=(BLOCK_DMODEL, BLOCK_N),
        order=(0, 1),
    )

    if IS_EVEN_N:
        k = tl.load(K_block_ptr)
    else:
        k = tl.load(K_block_ptr, boundary_check=(0, 1), padding_option="zero")


    lo = 0
    hi = ROUND_CTX
    qk_scale = sm_scale
    qk_scale *= 1.4426950408889634   # 1/log(2)

    for start_m in range(lo, hi, BLOCK_M):
        start_m = tl.multiple_of(start_m, BLOCK_M)
        if IS_EVEN_M:
            q = tl.load(Q_block_ptr)
        else:
            q = tl.load(Q_block_ptr, boundary_check=(0,1), padding_option="zero")

        m = tl.load(m_ptrs)

        # calc qk
        qk = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
        qk += tl.dot(q, k)
        qk = qk * qk_scale

        if SLIDING_WINDOW:
            dist = tl.arange(0, BLOCK_M)[:, None] - tl.arange(0, BLOCK_N)[None, :] \
                 + start_m - start_n * BLOCK_N + sliding_window_offset

            if COMPLEMENT_SLIDING_WINDOW:
                mask = (dist >= sliding_window_size)
            else:
                mask = (dist >= 0) & (dist < sliding_window_size)

        qk = qk - m[:, None]
        p = tl.math.exp2(qk) # (BLOCK_M, BLOCK_N)

        if SLIDING_WINDOW:
            p = tl.where(mask, p, 0)

        if not IS_EVEN_N:
            p = tl.where(
                ((tl.arange(0, BLOCK_M) + start_m) < N_CTX)[:, None],
                p, 0
            )

        o += tl.sum(p, axis=0)


        Q_block_ptr = tl.advance(Q_block_ptr, offsets=(BLOCK_M, 0))
        m_ptrs = m_ptrs + BLOCK_M

    o_offset = off_z.to(tl.int64) * stride_oz + off_h.to(tl.int64) * stride_oh
    o_range = tl.arange(0, BLOCK_N) + start_n * BLOCK_N # orange
    o_ptrs = Out + o_offset + o_range
    tl.store(o_ptrs, o.to(Out.type.element_ty), mask = o_range < NKV_CTX)

def get_score(q, k, m, sliding_window, complement_sliding_window):
    N_CTX = q.size(-2)
    NKV_CTX = k.size(-2)
    ROUND_CTX = m.size(-1)
    ret = torch.zeros(
        (q.size(0), q.size(1), k.size(2)),
        dtype=k.dtype, device=k.device
    )
    if sliding_window is not None:
        sliding_window_offset, sliding_window_size = sliding_window
    else:
        sliding_window_offset, sliding_window_size = None, None

    grid = lambda META: (
        triton.cdiv(k.shape[2], META["BLOCK_N"]),
        q.shape[0] * q.shape[1]
    )
    sm_scale = 1 / math.sqrt(q.size(-1))

    global _BLOCK_N
    global _BLOCK_M

    try:
        _score_kernel[grid](
            q, k, m, sm_scale, ret,
            q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            ret.stride(0), ret.stride(1), ret.stride(2),
            q.size(0), q.size(1), k.size(1),
            N_CTX, ROUND_CTX, NKV_CTX,
            sliding_window_offset,
            sliding_window_size,
            SLIDING_WINDOW=(sliding_window is not None),
            COMPLEMENT_SLIDING_WINDOW=complement_sliding_window,
            BLOCK_M=_BLOCK_M,
            BLOCK_N=_BLOCK_N,
            BLOCK_DMODEL=q.size(-1)
        )
    except triton.OutOfResources as E:
        from warnings import warn
        _BLOCK_N = _BLOCK_N // 2
        _BLOCK_M = _BLOCK_M // 2
        warn(f"Triton Attention Output Resources. {E}\nUse smaller block size {_BLOCK_N}.")
        _score_kernel[grid](
            q, k, m, sm_scale, ret,
            q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            ret.stride(0), ret.stride(1), ret.stride(2),
            q.size(0), q.size(1), k.size(1),
            N_CTX, ROUND_CTX, NKV_CTX,
            sliding_window_offset,
            sliding_window_size,
            SLIDING_WINDOW=(sliding_window is not None),
            COMPLEMENT_SLIDING_WINDOW=complement_sliding_window,
            BLOCK_M=_BLOCK_M,
            BLOCK_N=_BLOCK_N,
            BLOCK_DMODEL=q.size(-1)
        )

    return ret



##################################################################################################################################################


import torch

# Define the test function for get_score
def test_get_score():
    # Define input dimensions
    batch_size = 2
    num_heads = 4
    seq_len = 128
    d_model = 64

    # Create random input tensors
    q = torch.randn((batch_size, num_heads, seq_len, d_model), device='cuda', dtype=torch.float16)
    k = torch.randn((batch_size, num_heads, seq_len, d_model), device='cuda', dtype=torch.float16)
    m = torch.zeros((batch_size, num_heads, seq_len), device='cuda', dtype=torch.float32)

    # Define sliding window parameters
    sliding_window = (0, 64)
    complement_sliding_window = False

    # Call the get_score function
    ret1 = get_score(q, k, m, sliding_window, complement_sliding_window)

    # Test with complement_sliding_window = True
    complement_sliding_window = True
    ret2 = get_score(q, k, m, sliding_window, complement_sliding_window)

    # Test without sliding window
    sliding_window = None
    complement_sliding_window = False
    ret3 = get_score(q, k, m, sliding_window, complement_sliding_window)

    # Test with different sliding window size
    sliding_window = (0, 32)
    ret4 = get_score(q, k, m, sliding_window, complement_sliding_window)

    results = {
        "test_case_1": ret1,
        "test_case_2": ret2,
        "test_case_3": ret3,
        "test_case_4": ret4
    }
    return results

# Run the tests
result_gold = test_get_score()



# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.

import torch
import triton
import triton.language as tl
import sys
import argparse
import pytest
import re

# This is a Triton kernel for matrix multiplication (GEMM) with support for various data types and scaling modes.

#################### Helper utils functions ####################
# Activation function.  
@triton.jit  
def leaky_relu(x):  
    x = x + 1  
    return tl.where(x >= 0, x, 0.01 * x)  
#################### Helper utils functions ####################



@triton.autotune(  
    configs=[  
        triton.Config(  
            {  
                'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 4, 'waves_per_eu': 2,  
                'kpack': 2, 'matrix_instr_nonkdim': 16  
            }, num_warps=4, num_stages=2),  
        triton.Config(  
            {  
                'BLOCK_SIZE_M': 256, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 4, 'waves_per_eu': 2,  
                'kpack': 2, 'matrix_instr_nonkdim': 16  
            }, num_warps=8, num_stages=2),  
        triton.Config(  
            {'BLOCK_SIZE_M': 256, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 4, 'waves_per_eu': 0},  
            num_warps=8, num_stages=2),  
        triton.Config(  
            {  
                'BLOCK_SIZE_M': 256, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 4, 'waves_per_eu': 2,  
                'kpack': 1, 'matrix_instr_nonkdim': 16  
            }, num_warps=8, num_stages=2),  
        triton.Config(  
            {  
                'BLOCK_SIZE_M': 256, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 1, 'waves_per_eu': 0,  
                'kpack': 1  
            }, num_warps=8, num_stages=2),  
        triton.Config(  
            {'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 4, 'waves_per_eu': 0},  
            num_warps=8, num_stages=2),  
        triton.Config(  
            {'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 1, 'waves_per_eu': 2},  
            num_warps=8, num_stages=2),  
    ],  
    key=['M', 'N', 'K'],  
    use_cuda_graph=True,  
)  
@triton.heuristics({  
    'EVEN_K':  
    lambda args: args['K'] % args['BLOCK_SIZE_K'] == 0, 'GRID_MN':  
    lambda args: triton.cdiv(args['M'], args['BLOCK_SIZE_M']) * triton.cdiv(args['N'], args['BLOCK_SIZE_N'])  
})  
@triton.jit  
def matmul_kernel(  
    a_ptr,  
    b_ptr,  
    c_ptr,  
    M,  
    N,  
    K,  
    stride_am,  
    stride_ak,  
    stride_bk,  
    stride_bn,  
    stride_cm,  
    stride_cn,  
    a_scale_ptr,  
    b_scale_ptr,  
    stride_ascale_m,  
    stride_ascale_k,  
    stride_bscale_k,  
    stride_bscale_n,  
    # Meta-parameters  
    GROUP_K: tl.constexpr,  
    GROUP_N: tl.constexpr,  
    BLOCK_SIZE_M: tl.constexpr,  
    BLOCK_SIZE_N: tl.constexpr,  
    BLOCK_SIZE_K: tl.constexpr,  
    EVEN_K: tl.constexpr,  
    GROUP_SIZE_M: tl.constexpr,  
    APPLY_SCALE: tl.constexpr,  
    ACTIVATION: tl.constexpr,  
    GRID_MN: tl.constexpr,  
):  
    """Kernel for computing the matmul C = A x B.  
    A has shape (M, K), B has shape (K, N) and C has shape (M, N)  
    """  
  
    NUM_XCDS: tl.constexpr = 8  
  
    tl.static_assert(((APPLY_SCALE is None) or (APPLY_SCALE == 'tensor')) or (APPLY_SCALE == 'block'),  
                     f"Scaling mode {APPLY_SCALE} is not supported!!!")  
  
    tl.assume(stride_am > 0)  
    tl.assume(stride_ak > 0)  
    tl.assume(stride_bk > 0)  
    tl.assume(stride_bn > 0)  
    tl.assume(stride_cm > 0)  
    tl.assume(stride_cn > 0)  
    tl.assume(stride_ascale_m > 0)  
    tl.assume(stride_ascale_k > 0)  
    tl.assume(stride_bscale_k > 0)  
    tl.assume(stride_bscale_n > 0)  
  
    # -----------------------------------------------------------  
    # Map program ids `pid` to the block of C it should compute.  
    # This is done in a grouped ordering to promote L2 data reuse.  
    # TODO(vgokhale): Add XCD remapping.  
    pid = tl.program_id(axis=0)  
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)  
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)  
  
    ## pid remapping on xcds  
    # Number of pids per XCD in the new arrangement  
    pids_per_xcd = (GRID_MN + NUM_XCDS - 1) // NUM_XCDS  
    # When GRID_MN cannot divide NUM_XCDS, some xcds will have  
    # pids_per_xcd pids, the other will have pids_per_xcd - 1 pids.  
    # We calculate the number of xcds that have pids_per_xcd pids as  
    # tall_xcds  
    tall_xcds = GRID_MN % NUM_XCDS  
    tall_xcds = NUM_XCDS if tall_xcds == 0 else tall_xcds  
    # Compute current XCD and local pid within the XCD  
    xcd = pid % NUM_XCDS  
    local_pid = pid // NUM_XCDS  
    # Calculate new pid based on the new grouping  
    # Note that we need to consider the following two cases:  
    # 1. the current pid is on a tall xcd  
    # 2. the current pid is on a short xcd  
    if xcd < tall_xcds:  
        pid = xcd * pids_per_xcd + local_pid  
    else:  
        pid = tall_xcds * pids_per_xcd + (xcd - tall_xcds) * (pids_per_xcd - 1) + local_pid  
  
    if GROUP_SIZE_M == 1:  
        pid_m = pid // num_pid_n  
        pid_n = pid % num_pid_n  
    else:  
        num_pid_in_group = GROUP_SIZE_M * num_pid_n  
        group_id = pid // num_pid_in_group  
        first_pid_m = group_id * GROUP_SIZE_M  
        group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)  
        pid_m = first_pid_m + (pid % group_size_m)  
        pid_n = (pid % num_pid_in_group) // group_size_m  
  
    tl.assume(pid_m > 0)  
    tl.assume(pid_n > 0)  
  
    # Create pointers for first block of A and B input matrices  
    offs_k = tl.arange(0, BLOCK_SIZE_K)  
    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M  
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N  
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)  
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)  
    if APPLY_SCALE == 'tensor':  
        a_scale = tl.load(a_scale_ptr) if a_scale_ptr else 1.0  
        b_scale = tl.load(b_scale_ptr)  
    elif APPLY_SCALE == 'block':  
        k_start = 0  
        offs_ks = k_start // GROUP_K  
        a_scale_ptrs = None if a_scale_ptr is None else (a_scale_ptr + offs_am * stride_ascale_m +  
                                                         offs_ks * stride_ascale_k)  
        offs_bsn = offs_bn // GROUP_N  
        b_scale_ptrs = b_scale_ptr + offs_bsn * stride_bscale_n + offs_ks * stride_bscale_k  
  
    acc_dtype = tl.float32 if c_ptr.type.element_ty != tl.int8 else tl.int32  
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)  
  
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):  
        # Load the next block of A and B, generate a mask by checking the K dimension.  
        # If it is out of bounds, set it to 0.  
        if EVEN_K:  
            a = tl.load(a_ptrs)  
            b = tl.load(b_ptrs)  
        else:  
            a = tl.load(a_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)  
            b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K, other=0.0)  
  
        if APPLY_SCALE == 'block':  
            b_scale = tl.load(b_scale_ptrs)  
            if a_scale_ptrs is not None:  
                a_scale = tl.load(a_scale_ptrs)  
  
        # Type conversion to support mixed precision GEMMs where b is lower precision than a  
        b = b.to(a_ptr.type.element_ty)  
  
        if APPLY_SCALE == 'block':  
            if a_scale_ptrs is not None:  
                accumulator += tl.dot(a, b, input_precision="ieee") * a_scale[:, None] * b_scale[None, :]  
            else:  
                accumulator += tl.dot(a, b, input_precision="ieee") * b_scale[None, :]  
        else:  
            accumulator += tl.dot(a, b, input_precision="ieee")  
  
        # Advance the ptrs to the next K block.  
        a_ptrs += BLOCK_SIZE_K * stride_ak  
        b_ptrs += BLOCK_SIZE_K * stride_bk  
  
        if APPLY_SCALE == 'block':  
            k_cur = k * BLOCK_SIZE_K // GROUP_K  
            k_nxt = (k + 1) * BLOCK_SIZE_K // GROUP_K  
            offs_ks = k_nxt - k_cur  
            b_scale_ptrs += offs_ks * stride_bscale_k  
            if a_scale_ptrs is not None:  
                a_scale_ptrs += offs_ks * stride_ascale_k  
  
    # Apply scale to recover dynamic range reduced due to lower precision inputs.  
    if APPLY_SCALE == 'tensor':  
        accumulator = accumulator * a_scale * b_scale  
    # Apply activation function, if specified.  
    # TODO(vgokhale): Add different types of activations.  
    if ACTIVATION == "leaky_relu":  
        accumulator = leaky_relu(accumulator)  
    c = accumulator.to(c_ptr.type.element_ty)  
  
    # Write back the block of the output matrix C with masks.  
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)  
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)  
    c_ptrs = c_ptr + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]  
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)  
    tl.store(c_ptrs, c, mask=c_mask)  
  
  



##################################################################################################################################################

# This is a Triton kernel for matrix multiplication (GEMM) with support for various data types and scaling modes.
# Usage Instruction: python3 -m pytest gemm.py  
  
import torch  
import triton  
import triton.language as tl  
import sys  
import argparse  
import pytest  
import re  
from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict  
######################################## HELPERS for Eval ######################################## 
import numpy as np
import random
import torch 
import os

result_gold = {}
CONFIG = {
  "llama3": {
    "8B": {
      "num_attention_heads": 32,
      "num_key_value_heads": 8,
      "hidden_size": 4096,
      "intermediate_size": 14336,
      "vocab_size": 128256
    },
    "70B": {
      "num_attention_heads": 64,
      "num_key_value_heads": 8,
      "hidden_size": 8192,
      "intermediate_size": 28672,
      "vocab_size": 128256
    },
    "405B": {
      "num_attention_heads": 128,
      "num_key_value_heads": 8,
      "hidden_size": 16384,
      "intermediate_size": 53248,
      "vocab_size": 128256
    }
  },
  "mistral": {
    "7B": {
      "hidden_size": 4096,
      "intermediate_size": 14336,
      "num_attention_heads": 32,
      "num_key_value_heads": 8,
      "vocab_size": 32000
    },
    "22B": {
      "hidden_size": 6144,
      "intermediate_size": 16384,
      "num_attention_heads": 48,
      "num_key_value_heads": 8,
      "vocab_size": 32000
    }

  }
}

  
def get_model_configs(config_path='model_configs.json', model_families=["llama3"], model="all"):  
    """  
    Load model names from the configuration file.  
  
    Args:  
        config_path (str): User-provided path to the configuration JSON file.  
        model_families (list): List of model family names to retrieve.  
  
    Returns:  
        dict: A dictionary of available models and their configurations for the specified families.  
    """  
    configs = CONFIG.copy()
  
    # Extract models and their configurations for the specified families  
    filtered_configs = {}  
  
    for family in model_families:  
        if family in configs:  
            # Check if model filtering is required  
            if model == "all":  
                # Include all models in the family  
                for model_size, model_configs in configs[family].items():  
                    filtered_configs[f"{family}-{model_size}"] = model_configs  
            else:  
                # Parse the model string (e.g., llama3_8B or llama3-8B)  
                delimiter = "_" if "_" in model else "-"  
                model_parts = model.split(delimiter)  
  
                # Check if the family and size match  
                if len(model_parts) == 2 and model_parts[0] == family:  
                    model_size = model_parts[1]  
                    if model_size in configs[family]:  
                        filtered_configs[f"{family}-{model_size}"] = configs[family][model_size]  
  
    if not filtered_configs:  
        print(f"Warning: No models selected for families: {model_families} with filter: '{model}'")  
  
    return filtered_configs  
  
  
def get_available_models(config_file='model_configs.json', model_families=["llama3"]):  
    """  
    Load model names from the configuration file.  
  
    Args:  
        config_file (str): Path to the configuration JSON file.  
        model_families (list): List of model family names to retrieve.  
  
    Returns:  
        list: A list of available models for the specified families.  
    """  
    configs = CONFIG.copy()
  
    models = [f"{family}-{model}" for family in model_families if family in configs for model in configs[family]]  
  
    return models  


def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

# --- Define TFLOPS and GB/s calculators for GEMM ---
def calculate_gemm_tflops(params: dict, ms: float) -> float:
    M = params['M']
    N = params['N']
    K = params['K']
    # For GEMM: 2 * M * N * K FLOPs
    flops = 2 * M * N * K
    tflops = flops / (ms / 1000) / 1e12
    return tflops

def calculate_gemm_gbps(params: dict, ms: float) -> float:
    M = params['M']
    N = params['N']
    K = params['K']
    # Dtypes are needed for element size
    # Assuming params will contain 'in_dtype_a_str' and 'in_dtype_b_str'
    # or we can infer from the created tensors if passed differently.
    # For simplicity, let's assume fp16/bf16 for now if not specified.
    # A more robust way is to pass tensor objects or precise dtype info.
    try:
        dtype_a = name_to_torch_types[params['in_dtype_a_str']]
        dtype_b = name_to_torch_types[params['in_dtype_b_str']]
        # dtype_c = name_to_torch_types[params['out_dtype_str']] # If C is different
    except KeyError:
        print("Warning: Dtype strings not found in params for GB/s calc, assuming float16.")
        dtype_a = torch.float16
        dtype_b = torch.float16
        # dtype_c = torch.float16

    bytes_a = M * K * torch.tensor([], dtype=dtype_a).element_size()
    bytes_b = K * N * torch.tensor([], dtype=dtype_b).element_size()
    bytes_c = M * N * torch.tensor([], dtype=dtype_a).element_size() # Assuming C is same type as A for this calc

    # Read A, Read B, Write C
    total_bytes = bytes_a + bytes_b + bytes_c
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

######################################## HELPERS for Eval ######################################## 
   
SCALE_BLOCK_SIZE = 128  
  
  
# Wrapper for gemm kernel.  
def matmul(a, b, c, a_scale, b_scale, scale_a8_b8=None, activation=""):  
    # Check constraints.  
    assert a.shape[1] == b.shape[0], "Incompatible dimensions!!!"  
    assert (a.element_size()  
            >= b.element_size()), "Mixed dtype GEMMs are only supported when data type of a is bigger than b!!!"  
    assert (a.is_floating_point() == b.is_floating_point()  
            ), "GEMMs between float and integer type tensors are not supported!!!"  
    assert (scale_a8_b8 in [None, 'tensor', 'block']), f"Scaling mode {scale_a8_b8} is not supported!!!"  
    M, K = a.shape  
    K, N = b.shape  
    grid = lambda META: (triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']), )  
    matmul_kernel[grid](  
        a,  
        b,  
        c,  
        M,  
        N,  
        K,  
        a.stride(0),  
        a.stride(1),  
        b.stride(0),  
        b.stride(1),  
        c.stride(0),  
        c.stride(1),  
        a_scale,  
        b_scale,  
        a_scale.stride(0) if (a_scale is not None) and a_scale.ndim else 0,  
        a_scale.stride(1) if (a_scale is not None) and a_scale.ndim else 0,  
        b_scale.stride(0) if (b_scale is not None) and b_scale.ndim else 0,  
        b_scale.stride(1) if (b_scale is not None) and b_scale.ndim else 0,  
        GROUP_K=SCALE_BLOCK_SIZE,  
        GROUP_N=SCALE_BLOCK_SIZE,  
        APPLY_SCALE=scale_a8_b8,  
        ACTIVATION=activation,  
    )  
  
  
def is_cdna4():  
    return triton.runtime.driver.active.get_current_target().arch == 'gfx950'  
  
  
e5m2_type = torch.float8_e5m2 if is_cdna4() else torch.float8_e5m2fnuz  
e4m3_type = torch.float8_e4m3fn if is_cdna4() else torch.float8_e4m3fnuz  
  
name_to_torch_types = {  
    'int8': torch.int8,  
    'int32': torch.int32,  
    'fp16': torch.float16,  
    'fp32': torch.float32,  
    'bf16': torch.bfloat16,  
    'fp8e5': e5m2_type,  
    'fp8e4': e4m3_type,  
}  
  
dtype_max = {  
    dtype: (torch.finfo(dtype) if dtype.is_floating_point else torch.iinfo(dtype)).max  
    for dtype in [  
        e5m2_type,  
        e4m3_type,  
        torch.int8,  
    ]  
}  
  
  
def dtype_is_8_bit(dtype):
    return (
        dtype is e5m2_type
        or dtype is e4m3_type
        or dtype is torch.int8
    )
  
  
def gen_input(M, N, dtype, needTrans, seed=0, fp8_scaling_mode='tensor', device='cuda'):  
    set_seed()  
  
    if needTrans:  
        raw_data = torch.randn((N, M), dtype=torch.float32, device='cuda').T  
    else:  
        raw_data = torch.randn((M, N), dtype=torch.float32, device='cuda')  
    scale = None  
    if dtype_is_8_bit(dtype):  
        if fp8_scaling_mode == 'token':  
            assert raw_data.size(1) % SCALE_BLOCK_SIZE == 0  
            raw_data = raw_data.view(M, -1, SCALE_BLOCK_SIZE)  
            max_val = raw_data.abs().float().amax(dim=2).view(M, -1).clamp(1e-4)  
            scale = max_val.unsqueeze(2) / dtype_max[dtype]  
            raw_data = (raw_data / scale).view(M, N)  
            scale = scale.view(M, -1)  
            scale = scale.T.contiguous().T  
        elif fp8_scaling_mode == 'block':  
            x_padded = torch.zeros((triton.cdiv(M, SCALE_BLOCK_SIZE) * SCALE_BLOCK_SIZE,  
                                    triton.cdiv(N, SCALE_BLOCK_SIZE) * SCALE_BLOCK_SIZE), dtype=raw_data.dtype,  
                                   device=raw_data.device)  
            x_padded[:M, :N] = raw_data  
            x_view = x_padded.view(-1, SCALE_BLOCK_SIZE, x_padded.size(1) // SCALE_BLOCK_SIZE, SCALE_BLOCK_SIZE)  
            x_amax = x_view.abs().float().amax(dim=(1, 3), keepdim=True).clamp(1e-4)  
            x_scaled = x_view * (dtype_max[dtype] / x_amax)  
            raw_data = x_scaled.view_as(x_padded)[:M, :N].T.contiguous().T  
            scale = (x_amax / dtype_max[dtype]).view(x_view.size(0), x_view.size(2))  
        elif fp8_scaling_mode == 'tensor':  
            max_val = torch.max(torch.abs(raw_data))  
            scale = max_val / dtype_max[dtype]  
            raw_data = raw_data / scale  
  
    input = raw_data.to(dtype)  
    input_f32 = input.to(torch.float32)  
  
    return input, input_f32, scale  
  
  
def get_x_vals():  
    x_vals = [(1024 * v, 1024 * v, 1024 * v) for v in range(1, 9)]  
  
    x_vals += [(4864, 4096, 8192), (9728, 8192, 65536), (4864, 8192, 4160)]  
  
    return x_vals  
  
  
# Unit tests  
#TODO(vgokhale): Test activation.  
# yapf: disable  
@pytest.mark.parametrize(  
    "M, N, K, in_dtype_a, in_dtype_b, out_dtype, col_a, col_b",  
    [(*shape, in_dtype_a, in_dtype_b, out_dtype, col_a, col_b)  
     for shape in get_x_vals()  
     for in_dtype_a, in_dtype_b, out_dtype in [  
        ('fp16', 'fp16', 'fp16'),   ('bf16', 'bf16', 'bf16'),   ('fp32', 'fp32', 'fp32'),  
        ('fp8e4', 'fp8e4', 'fp16'), ('fp8e5', 'fp8e5', 'fp16'), ('fp16', 'fp8e4', 'fp16'),  
        ('fp16', 'fp8e5', 'fp16'),  ('bf16', 'fp8e4', 'bf16'),  ('bf16', 'fp8e5', 'bf16'),  
        ('int8', 'int8', 'int8'),   ('int8', 'int8', 'int32')]  
     # Defines if a matrix is row or column major.  
     for col_a in [True, False]  
     for col_b in [True, False]])  
# yapf: enable  
def test_correctness(M, N, K, col_a, col_b, in_dtype_a, in_dtype_b, out_dtype, request):  
    set_seed()
    torch_in_dtype_a = name_to_torch_types[in_dtype_a]  
    torch_in_dtype_b = name_to_torch_types[in_dtype_b]  
    a, a_fp32, a_scale = gen_input(M, K, torch_in_dtype_a, col_a, seed=1, device='cuda')  
    b, b_fp32, b_scale = gen_input(K, N, torch_in_dtype_b, col_b, seed=2, device='cuda')  
    torch_out_dtype = name_to_torch_types[out_dtype]  
    c = torch.empty((M, N), device=a.device, dtype=torch_out_dtype)  
    # For 8-bit, we have scaled to the dynamic range of the data type.  
    # This requires us to compute in fp32 because for e5m2, the range is same as fp16 (e5m10).  
    # If we use fp16 it is possible to return infs from the torch.matmul call.  
    if dtype_is_8_bit(torch_in_dtype_a) or dtype_is_8_bit(torch_in_dtype_b):  
        matmul(a, b, c, a_scale, b_scale, scale_a8_b8='tensor', activation="")  
        torch_output = torch.matmul(a_fp32, b_fp32)  
        # Set a_scale to 1.0 if it is not set  
        torch_output = torch_output * (a_scale or 1.0) * b_scale  
    # For other dtypes, use the same torch matmul as the dtype.  
    else:  
        matmul(a, b, c, a_scale=None, b_scale=None, scale_a8_b8=None, activation="")  
        torch_output = torch.matmul(a.to(torch_in_dtype_a), b.to(torch_in_dtype_b))  
    
    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])

    if out_dtype == 'int8':  
        ################### save c in result_gold ###################
        test_case_name = request.node.name
        sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
        result_gold[sanitized_key_name] = c.to(torch.float32).clone().detach().cpu()
        ###################################################################
        # torch.testing.assert_close(c.to(torch.float32),  
        #                            torch_output.to(torch.int8).to(torch.float32), atol=1e-3, rtol=1e-2)  
    else:  
        ################### save c in result_gold ###################
        test_case_name = request.node.name
        sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
        result_gold[sanitized_key_name] = c.clone().detach().cpu()
        ###################################################################
        # torch.testing.assert_close(c, torch_output.to(torch_out_dtype), atol=5e-3, rtol=1e-2)  
  
OP_NAME_FOR_BENCHMARK = "gemm_triton_perf"

@pytest.mark.parametrize(
    "M, N, K, in_dtype_a_str, in_dtype_b_str, out_dtype_str, col_a, col_b",
    [(*shape, ida, idb, od, ca, cb)
     for shape in get_x_vals() # Uses the reduced get_x_vals for faster testing
     for ida, idb, od in [
        ('fp16', 'fp16', 'fp16'), ('bf16', 'bf16', 'bf16'), ('fp32', 'fp32', 'fp32'),
        # ('fp8e4', 'fp8e4', 'fp16'), ('fp8e5', 'fp8e5', 'fp16'), # FP8 needs careful scale handling
        # ('int8', 'int8', 'int32') # Int8 also needs care
        ]
     for ca in [False] # Simplified: only row-major A for now
     for cb in [False]] # Simplified: only row-major B for now
)
def test_performance(M, N, K, col_a, col_b, in_dtype_a_str, in_dtype_b_str, out_dtype_str, request):
    set_seed() # Consistent seed for input data generation
    torch_in_dtype_a = name_to_torch_types[in_dtype_a_str]
    torch_in_dtype_b = name_to_torch_types[in_dtype_b_str]
    torch_out_dtype = name_to_torch_types[out_dtype_str]

    # --- Input Generation (from original test_correctness) ---
    # Determine fp8_scaling_mode based on dtypes, or make it a parameter
    fp8_mode_a = 'tensor' if dtype_is_8_bit(torch_in_dtype_a) else None
    fp8_mode_b = 'tensor' if dtype_is_8_bit(torch_in_dtype_b) else None

    a, a_fp32_ref, a_scale = gen_input(M, K, torch_in_dtype_a, col_a, seed=1, fp8_scaling_mode=fp8_mode_a or 'tensor', device='cuda')
    b, b_fp32_ref, b_scale = gen_input(K, N, torch_in_dtype_b, col_b, seed=2, fp8_scaling_mode=fp8_mode_b or 'tensor', device='cuda')
    c = torch.empty((M, N), device=a.device, dtype=torch_out_dtype)

    # Determine scale_a8_b8 and activation for matmul wrapper
    current_scale_a8_b8 = None
    current_activation = "" # No activation for pure GEMM perf
    if dtype_is_8_bit(torch_in_dtype_a) or dtype_is_8_bit(torch_in_dtype_b):
        current_scale_a8_b8 = 'tensor' # Or make this part of parametrize if testing block/token scaling perf

    # --- Create op_lambda for benchmarking ---
    # This lambda will call the matmul wrapper
    op_lambda = lambda: matmul(a, b, c, a_scale, b_scale,
                               scale_a8_b8=current_scale_a8_b8,
                               activation=current_activation)

    # --- Benchmarking ---
    bench_config = do_bench_config(warm_up=10, repetition=100) # Adjust for GEMM complexity
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)

    current_params_for_logs_and_calc = {
        "M": M, "N": N, "K": K,
        "in_dtype_a_str": in_dtype_a_str, "in_dtype_b_str": in_dtype_b_str, "out_dtype_str": out_dtype_str,
        "col_a": col_a, "col_b": col_b,
        "fp8_scaling_a": fp8_mode_a, "fp8_scaling_b": fp8_mode_b,
        "triton_scale_mode": current_scale_a8_b8, "activation": current_activation
    }

    benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                              gbps_calculator=calculate_gemm_gbps,
                              tflops_calculator=calculate_gemm_tflops)
  
def get_type(provider):  
    res = re.findall(r'\(.*?\)', provider)  
    return res[0][1:-1].split('/', 1)  
  
  
@triton.testing.perf_report(  
    triton.testing.Benchmark(  
        x_names=['M', 'N', 'K'],  
        x_vals=get_x_vals(),  
        line_arg='provider',  
        line_vals=[  
            'hipblaslt(fp16/fp16)', 'hipblaslt(bf16/bf16)', 'triton(fp16/fp16)', 'triton(bf16/bf16)',  
            'triton(int8/int8)', 'triton(fp8e4/fp8e4)', 'triton(fp8e5/fp8e5)', 'triton(fp16/fp8e4)',  
            'triton(fp16/fp8e5)'  
        ],  
        line_names=[  
            "rocBLAS.Fp16", "rocBLAS.Bf16", "Triton.Fp16", "Triton.Bf16", "Triton.Int8", "Triton.Fp8E4", "Triton.Fp8E5",  
            "Triton.Fp16.Fp8E4", "Triton.Fp16.Fp8E5"  
        ],  
        ylabel="TFLOPS",  
        plot_name="matmul-performance",  
        args={},  
    ))  
def benchmark(M, N, K, provider, model=None, args=None):  
    in_dtype_a, in_dtype_b = [name_to_torch_types[x] for x in get_type(provider)]  
    out_dtype = in_dtype_a  
  
    quantiles = [0.5, 0.2, 0.8]  
    layout_tn = args.layout == 'tn'  
  
    if args.fp8_scaling_mode == 'tensor' or in_dtype_b == torch.int8:  
        a, _, a_scale = gen_input(M, K, in_dtype_a, False, seed=1, device='cuda')  
        b, _, b_scale = gen_input(K, N, in_dtype_b, layout_tn, seed=2, device='cuda')  
    else:  
        a, _, a_scale = gen_input(M, K, in_dtype_a, False, seed=1, fp8_scaling_mode='token', device='cuda')  
        b, _, b_scale = gen_input(K, N, in_dtype_b, layout_tn, seed=2, fp8_scaling_mode='block', device='cuda')  
  
    if 'hipblaslt' in provider:  
        ms, min_ms, max_ms = triton.testing.do_bench(lambda: torch.matmul(a, b), quantiles=quantiles)  
    else:  # triton, different data types  
        assert "triton" in provider  
        # Allocates output.  
        c = torch.empty((M, N), device=a.device, dtype=out_dtype)  
  
        # If data type is 8 bit  
        #   Default to tensor scaling if scaling mode is tensor or dtype is int8  
        #   Use block scaling otherwise  
        scale_a8_b8 = None  
        if dtype_is_8_bit(in_dtype_a) or dtype_is_8_bit(in_dtype_b):  
            scale_a8_b8 = 'tensor' if in_dtype_b == torch.int8 else args.fp8_scaling_mode  
  
        ms, min_ms, max_ms = triton.testing.do_bench(  
            lambda: matmul(a, b, c, a_scale, b_scale, scale_a8_b8=scale_a8_b8, activation=""), quantiles=quantiles)  
        if args.v:  
            print(f'Best tuning config for M={M}, N={N}, K={K}, '  
                  f'dtype={in_dtype_a} / {in_dtype_b} / {out_dtype}: \n({matmul_kernel.best_config})\n')  
    perf = lambda ms: 2 * M * N * K * 1e-12 / (ms * 1e-3)  
    return perf(ms), perf(max_ms), perf(min_ms)  
  

######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all c_triton_dot results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} c_triton_dot tensors to {OUTPUT_FILENAME}.")  

def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")
######################################## HELPERS for Eval ######################################## 
  
def parse_args():  
    parser = argparse.ArgumentParser(  
        prog="AMD Triton GEMM kernel",  
        allow_abbrev=False,  
    )  
  
    parser.add_argument('-model_configs', type=str, default="model_configs.json", help="Model config json file.")  
  
    available_models = get_available_models(model_families=["llama3"])  # Dynamically load model names  
    model_help = (  
        "Model name to benchmark. Select from: [" + ", ".join(available_models) +  
        "]. Use 'all' to benchmark all models. Not providing runs the default benchmark script with custom configs.")  
    parser.add_argument('-model', type=str, default=None, help=model_help)  
    parser.add_argument("-v", action='store_true', default=False, help="Print out the best tuning config")  
    parser.add_argument("-M", type=int, default=0)  
    parser.add_argument("-N", type=int, default=0)  
    parser.add_argument("-K", type=int, default=0)  
    parser.add_argument("-layout", type=str, default='tn')  
    parser.add_argument("-dtype", type=str, default=None, help="Data type of inputs and outputs")  
    parser.add_argument("-b_dtype", type=str, default=None,  
                        help="Data type of B operand, if specified (else same as dtype)")  
    parser.add_argument("-fp8_scaling_mode", type=str, default='tensor', choices=['tensor', 'block'],  
                        help="Type of scaling to apply when either or both inputs are fp8")  
  
    args = parser.parse_args()  
  
    return args  
  
  
def get_line_vals_names(a_dtype=None, b_dtype=None):  
    line_vals = [  
        'hipblaslt(fp16/fp16)', 'hipblaslt(bf16/bf16)', 'triton(fp16/fp16)', 'triton(bf16/bf16)', 'triton(int8/int8)',  
        'triton(fp8e4/fp8e4)', 'triton(fp8e5/fp8e5)', 'triton(fp16/fp8e4)', 'triton(fp16/fp8e5)'  
    ]  
    line_names = [  
        "rocBLAS.Fp16", "rocBLAS.Bf16", "Triton.Fp16", "Triton.Bf16", "Triton.Int8", "Triton.Fp8E4", "Triton.Fp8E5",  
        "Triton.Fp16.Fp8E4", "Triton.Fp16.Fp8E5"  
    ]  
    assert not ((a_dtype is None) ^ (b_dtype is None))  
    if a_dtype is not None:  
        line_vals_suffix_str = '(' + a_dtype + '/' + b_dtype + ')'  
        line_names_suffix_str = '.' + a_dtype + '.' + b_dtype  
        line_vals = ['triton' + line_vals_suffix_str]  
        line_names = ['Triton' + line_names_suffix_str]  
        if (not dtype_is_8_bit(name_to_torch_types[a_dtype])
            and not dtype_is_8_bit(name_to_torch_types[b_dtype])):
            line_vals += ['hipblaslt' + line_vals_suffix_str]
            line_names += ['hipblaslt' + line_names_suffix_str]
  
    return line_vals, line_names  
  
  
def main():  
    args = parse_args()  
  
    if args.model:  
        config_file = args.model_configs  
        configs = get_model_configs(config_path=config_file, model_families=["llama3"], model=args.model)  
        mnk_list = []  
  
        for model_name, config in configs.items():  
            M, N, K = args.M or 8192, config["hidden_size"], config["intermediate_size"]  
            mnk_list.append((model_name, M, N, K))  
  
        benchmark.benchmarks.x_names = ['model', 'M', 'N', 'K']  
        benchmark.benchmarks.x_vals = mnk_list  
  
    a_dtype = args.dtype  
    b_dtype = args.b_dtype or args.dtype  
    assert a_dtype is None or a_dtype in name_to_torch_types, f"Unsupported dtype {a_dtype}"  
    assert b_dtype is None or b_dtype in name_to_torch_types, f"Unsupported dtype {b_dtype}"  
    benchmark.benchmarks.line_vals, benchmark.benchmarks.line_names = get_line_vals_names(a_dtype, b_dtype)  
    if args.N or args.K:  
        assert args.model is None, "Providing both -model and N/K is not compatible! -model already fixes N/K."  
  
    if args.M and args.N and args.K:  
        x_vals = [(args.M, args.N, args.K)]  
        benchmark.benchmarks.x_vals = x_vals  
  
    benchmark.run(show_plots=True, print_data=True, args=args)  
  
  
if __name__ == '__main__':  
    sys.exit(main())  
